# -*- coding: utf-8 -*-
"""
Input plug-in for a voronoi calculation.
"""
from aiida.engine import CalcJob
from aiida.orm import CalcJobNode, Dict, StructureData, RemoteData, SinglefileData
from aiida.common.utils import classproperty
from aiida.common.exceptions import (InputValidationError, ValidationError)
from aiida.common.datastructures import (CalcInfo, CodeInfo)
from aiida_kkr.tools.common_workfunctions import generate_inputcard_from_structure, check_2Dinput_consistency, vca_check
from aiida.common.exceptions import UniquenessError
import os

__copyright__ = (u'Copyright (c), 2017, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.5.4'
__contributors__ = ('Jens Broeder', 'Philipp Rüßmann')


class VoronoiCalculation(CalcJob):
    """
    AiiDA calculation plugin for a voronoi calculation (creation of starting potential and shapefun).
    """

    ####################
    # File names etc.
    ####################
    # calculation plugin version
    _CALCULATION_PLUGIN_VERSION = __version__
    # Default input and output files
    _DEFAULT_INPUT_FILE = 'inputcard'  # will be shown with inputcat
    _DEFAULT_OUTPUT_FILE = 'out_voronoi'  #'shell output will be shown with outputca
    # List of mandatory input files
    _INPUT_FILE_NAME = 'inputcard'
    # List of output files that should always be present
    _OUTPUT_FILE_NAME = 'out_voronoi'
    # template.product entry point defined in setup.json
    _default_parser = 'kkr.voroparser'
    # File names
    _ATOMINFO = 'atominfo.txt'
    _RADII = 'radii.dat'
    _SHAPEFUN = 'shapefun'
    _VERTICES = 'vertices.dat'
    _OUT_POTENTIAL_voronoi = 'output.pot'
    _POTENTIAL_IN_OVERWRITE = 'overwrite_potential'

    @classmethod
    def define(cls, spec):
        """
        define internals and inputs / outputs of calculation
        """
        # reuse base class (i.e. CalcJob) functions
        super(VoronoiCalculation, cls).define(spec)
        # now define input files and parser
        spec.input('metadata.options.parser_name', valid_type=str, default=cls._default_parser, non_db=True)
        spec.input('metadata.options.input_filename', valid_type=str, default=cls._DEFAULT_INPUT_FILE, non_db=True)
        spec.input('metadata.options.output_filename', valid_type=str, default=cls._DEFAULT_OUTPUT_FILE, non_db=True)
        # define input nodes (optional ones have required=False)
        spec.input('parameters', valid_type=Dict, help='Use a node that specifies the input parameters')
        spec.input(
            'structure',
            valid_type=StructureData,
            required=False,
            help='Use a node that specifies the input crystal structure'
        )
        spec.input(
            'parent_KKR',
            valid_type=RemoteData,
            required=False,
            help='Use a node that specifies a parent KKR calculation'
        )
        spec.input(
            'potential_overwrite',
            valid_type=SinglefileData,
            required=False,
            help='Use a node that specifies the potential which is used instead of the voronoi output potential'
        )
        spec.input(
            'shapefun_overwrite',
            valid_type=SinglefileData,
            required=False,
            help='Use a node that specifies the shapefun which is used instead of the voronoi output'
        )
        # define outputs
        spec.output('output_parameters', valid_type=Dict, required=True, help='results of the calculation')
        spec.default_output_node = 'output_parameters'
        # define exit codes, also used in parser
        spec.exit_code(301, 'ERROR_NO_OUTPUT_FILE', message='Voronoi output file not found')
        spec.exit_code(302, 'ERROR_VORONOI_PARSING_FAILED', message='Voronoi parser retuned an error')

    def prepare_for_submission(self, tempfolder):
        """Create the input files from the input nodes passed to this instance of the `CalcJob`.

        :param tempfolder: an `aiida.common.folders.Folder` to temporarily write files on disk
        :return: `aiida.common.datastructures.CalcInfo` instance
        """
        # Check inputdict
        parameters = self.inputs.parameters

        if 'structure' in self.inputs:
            structure = self.inputs.structure
            found_structure = True
        else:
            found_structure = False

        vca_structure = False
        if found_structure:
            # for VCA: check if input structure and parameter node define VCA structure
            vca_structure = vca_check(structure, parameters)  # pylint: disable=used-before-assignment

        code = self.inputs.code

        # check if a parent folder containing a potential file (out_potential) is given
        if 'parent_KKR' in self.inputs:
            parent_calc_folder = self.inputs.parent_KKR
            found_parent = True
        else:
            found_parent = False

        if found_parent:
            # check if parent is either Voronoi or previous KKR calculation
            overwrite_potential, parent_calc = self._check_valid_parent(parent_calc_folder)  # pylint: disable=possibly-used-before-assignment

            #cross check if no structure was given and extract structure from parent
            if found_structure and not vca_structure:
                raise InputValidationError(
                    'parent_KKR and structure found in input. '
                    'Can only use either parent_KKR or structure in input.'
                )
            else:
                structure_remote_KKR, voro_parent = self.find_parent_structure(parent_calc)
                if not vca_structure:
                    structure = structure_remote_KKR
                else:
                    # check consistency of input vca structure and structure  from remote KKR folder
                    # TODO check consistency
                    pass
        else:
            overwrite_potential = False
            if not found_structure:
                raise InputValidationError('Neither structure nor parent_KKR specified for this '
                                           'calculation')

        # check if overwrite potential is given explicitly
        if 'potential_overwrite' in self.inputs:
            potfile_overwrite = self.inputs.potential_overwrite
            has_potfile_overwrite = True
        else:
            has_potfile_overwrite = False

        if has_potfile_overwrite:
            overwrite_potential = True
            if not found_structure:
                raise InputValidationError(
                    'Input structure needed for this calculation '
                    "(using 'potential_overwrite' input node)"
                )

        ###################################
        # Check for 2D case
        twoDimcheck, msg = check_2Dinput_consistency(structure, parameters)
        if not twoDimcheck:
            raise InputValidationError(msg)

        # Prepare inputcard from Structure and input parameter data
        with tempfolder.open(self._INPUT_FILE_NAME, u'w') as input_file:
            try:
                use_alat_input = parameters.get_dict().get('use_input_alat', False)
                use_alat_input = parameters.get_dict().get('USE_INPUT_ALAT', use_alat_input)
                natom, nspin, newsosol, warnings_write_inputcard = generate_inputcard_from_structure(
                    parameters,
                    structure,
                    input_file,
                    isvoronoi=True,
                    vca_structure=vca_structure,
                    use_input_alat=use_alat_input,
                )
            except ValueError as e:
                raise InputValidationError(f'Input Dict not consistent: {e}')

        # Decide what files to copy
        local_copy_list = []
        if overwrite_potential:
            # copy the right files
            #TODO check first if file, exists and throw
            # warning, now this will throw an error
            if found_parent and self._is_KkrCalc(parent_calc):
                outfolder = parent_calc.outputs.retrieved  # copy from remote folder
                copylist = [parent_calc.process_class._OUT_POTENTIAL]
            elif has_potfile_overwrite:
                outfolder = potfile_overwrite  # copy from potential sfd  # pylint: disable=possibly-used-before-assignment
                copylist = [potfile_overwrite.filename]
            else:
                copylist = []

            for file1 in copylist:
                filename = file1
                if (found_parent or has_potfile_overwrite) and file1 == copylist[0]:
                    filename = self._POTENTIAL_IN_OVERWRITE
                local_copy_list.append((outfolder.uuid, file1, filename))  # pylint: disable=possibly-used-before-assignment

        # add shapefun to overwrite
        if 'shapefun_overwrite' in self.inputs:
            shapefun_overwrite = self.inputs.shapefun_overwrite
            filename = shapefun_overwrite.filename
            local_copy_list.append((shapefun_overwrite.uuid, filename, 'shapefun_overwrite'))

        # Prepare CalcInfo to be returned to aiida
        calcinfo = CalcInfo()
        calcinfo.uuid = self.uuid
        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = []
        calcinfo.retrieve_list = [
            self._OUTPUT_FILE_NAME,
            self._ATOMINFO,
            self._RADII,
            self._SHAPEFUN,
            self._VERTICES,
            self._INPUT_FILE_NAME,
        ]

        # pass on overwrite potential if this was given in input
        # (KkrCalculation checks if this file is there and takes this file instead of _OUT_POTENTIAL_voronoi
        #  if given)
        if overwrite_potential:
            calcinfo.retrieve_list += [self._POTENTIAL_IN_OVERWRITE]
        else:
            calcinfo.retrieve_list += [self._OUT_POTENTIAL_voronoi]

        codeinfo = CodeInfo()
        codeinfo.cmdline_params = []
        codeinfo.stdout_name = self._OUTPUT_FILE_NAME
        codeinfo.code_uuid = code.uuid
        calcinfo.codes_info = [codeinfo]

        return calcinfo

    def _check_valid_parent(self, parent_calc_folder):
        """
        Check that calc is a valid parent for a FleurCalculation.
        It can be a VoronoiCalculation, KKRCalculation
        """
        overwrite_pot = False

        # extract parent calculation
        parent_calcs = parent_calc_folder.get_incoming(node_class=CalcJobNode)
        n_parents = len(parent_calcs.all_link_labels())
        if n_parents != 1:
            raise UniquenessError(
                'Input RemoteData is child of {} '
                'calculation{}, while it should have a single parent'
                ''.format(n_parents, '' if n_parents == 0 else 's')
            )
        else:
            parent_calc = parent_calcs.first().node
            overwrite_pot = True

        if ((not self._is_KkrCalc(parent_calc))):
            raise ValueError('Parent calculation must be a KkrCalculation')

        return overwrite_pot, parent_calc

    def _is_KkrCalc(self, calc):
        """
        check if calc contains the file out_potential
        """
        is_KKR = False
        if calc.process_type == 'aiida.calculations:kkr.kkr':
            retrieved_node = calc.get_retrieved_node()
            if 'out_potential' in retrieved_node.list_object_names():
                is_KKR = True

        return is_KKR

    @classmethod
    def find_parent_structure(self, parent_folder):
        """
        Find the Structure node recuresively in chain of parent calculations (structure node is input to voronoi calculation)
        This is a copy of the find_parent_structure that moved to tools.find_parent to keep backwards compatibility.
        """
        from aiida_kkr.tools.find_parent import find_parent_structure
        return find_parent_structure(parent_folder)



"""
Calculations provided by plugin

Register calculations via the "aiida.calculations" entry point in setup.json.
"""

# import all calculations here to expose them in `aiida_kkr.calculations` directly
from .voro import VoronoiCalculation
from .kkr import KkrCalculation
from .kkrimp import KkrimpCalculation
# broken at the moment:
#from .kkrimporter import KkrimporterCalculation



# -*- coding: utf-8 -*-
"""
In this module you find the total workflow for a kkr impurity calculation
and some helper methods to do so with AiiDA
"""

from aiida.orm import Code, load_node, RemoteData, StructureData, Dict, SinglefileData, FolderData
from aiida.orm import CalcJobNode
from aiida.engine import WorkChain, ToContext, if_
from aiida.engine import calcfunction
from aiida_kkr.calculations.voro import VoronoiCalculation
from masci_tools.io.kkr_params import kkrparams
from aiida_kkr.tools import test_and_get_codenode, neworder_potential_wf, update_params_wf
from aiida_kkr.workflows.gf_writeout import kkr_flex_wc
from aiida_kkr.workflows.voro_start import kkr_startpot_wc
from aiida_kkr.workflows.kkr_imp_sub import kkr_imp_sub_wc
import numpy as np
from aiida_kkr.tools.save_output_nodes import create_out_dict_node

__copyright__ = (u'Copyright (c), 2017, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.9.3'
__contributors__ = (u'Fabian Bertoldo', u'Philipp Rüßmann')
#TODO: generalize workflow to multiple impurities
#TODO: add additional checks for the input
#TODO: maybe work on a clearer outputnode structure


class kkr_imp_wc(WorkChain):
    """
    Workchain of a kkrimp calculation starting either from scratch (with a structure
    and impurity_info node), or with a converged host potential and impurity
    startpotentials, ... to calculate the converged host-impurity potential of the system.

    :param options: (Dict), Workchain specifications
    :param wf_parameters: (Dict), specifications for the kkr impurity workflow
    :param voro_aux_parameters: (Dict), specification for the auxiliary voronoi calculation for the impurity
    :param kkrimp: (Code), mandatory: KKRimp code converging the host-imp-potential
    :param kkr: (Code), mandatory: KKR code for calculation the host potential
    :param voronoi: (Code), mandatory: Voronoi code to generate the impurity startpot
    :param remote_data_gf: (RemoteData): remote folder of a previous kkrflex
                                         calculation containing the flexfiles ...
    :param remote_data_host: (RemoteData): remote folder of a converged KKR
                                           host calculation

    :return workflow_info: (Dict), Information of workflow results
    :return last_calc_output_parameters: (Dict), output parameters of
                                         the last called calculation
    :return last_calc_info: (Dict), information of the last called calculation
    """

    _workflowversion = __version__
    _wf_label = 'kkr_imp_wc'
    _wf_description = 'Workflow for a KKRimp calculation'

    _options_default = {
        'queue_name': '',  # Queue name to submit jobs too
        'resources': {
            'num_machines': 1  # resources to allowcate for the job
        },
        'max_wallclock_seconds': 60 * 60,  # walltime after which the job gets killed (gets parsed to KKR)}
        'custom_scheduler_commands': '',  # some additional scheduler commands
        'withmpi': True  # execute KKR with mpi or without
    }

    # settings for sub workflow (impurity convergence)
    _wf_default = kkr_imp_sub_wc.get_wf_defaults(silent=True)
    # add control to retrieve kkrflex files to repository or leave on remote computer only
    _wf_default['retrieve_kkrflex'] = True

    # settings for vorostart workflow, used to generate starting potential
    _voro_aux_default = kkr_startpot_wc.get_wf_defaults(silent=True)

    @classmethod
    def get_wf_defaults(self, silent=False):
        """
        Print and return _wf_defaults dictionary. Can be used to easily create
        set of wf_parameters.

        returns _wf_defaults
        """
        if not silent:
            print(f'Version of workflow: {self._workflowversion}')
        return self._options_default.copy(), self._wf_default.copy(), self._voro_aux_default.copy()

    @classmethod
    def define(cls, spec):
        """
        Defines the outline of the workflow
        """

        super(kkr_imp_wc, cls).define(spec)

        # expose inputs of sub workflow
        # TODO also expose the other inputs in next release, but put deprecation warnings in first
        spec.expose_inputs(
            kkr_imp_sub_wc,
            namespace='scf',
            include=(
                # 'kkrimp',
                'options',
                # 'wf_parameters',
                'params_overwrite',
                'initial_noco_angles',
                'rimpshift',
                'settings_LDAU',
            )
        )

        # define the inputs of the workflow
        spec.input('kkr', valid_type=Code, required=False, help='KKRhost code used to run GF writeout step.')
        spec.input(
            'voronoi',
            valid_type=Code,
            required=True,
            help='Voronoi code used to create the impurity starting potential.'
        )
        spec.input(
            'kkrimp', valid_type=Code, required=True, help='KKRimp code used to converge the impurity calculation'
        )
        spec.input(
            'impurity_info',
            valid_type=Dict,
            required=True,
            help='Information of the impurity like position in the unit cell, screening cluster, atom type.'
        )
        spec.input(
            'remote_data_host',
            valid_type=RemoteData,
            required=False,
            help='RemoteData node of the converged host calculation. Used to write out the host Green function.'
        )
        spec.input(
            'remote_data_gf',
            valid_type=RemoteData,
            required=False,
            help='RemoteData node of precomputed host Green function.'
        )
        spec.input(
            'remote_data_gf_Efshift',
            valid_type=RemoteData,
            required=False,
            help=
            'RemoteData node of precomputed host Green function with Fermi level shift (overwrite kkrflex_green and tmat files from first remote_data_gf node.'
        )
        spec.input('options', valid_type=Dict, required=False, help='Options for running the codes (walltime etc.).')
        spec.input(
            'options_voronoi',
            valid_type=Dict,
            required=False,
            help='Options for running the Voronoi code (if differing from general `options` node)'
        )
        spec.input(
            'voro_aux_parameters',
            valid_type=Dict,
            required=False,
            help='Parameters for the auxiliary voronoi starting potential workflow.'
        )
        spec.input(
            'wf_parameters',
            valid_type=Dict,
            required=False,
            help='Parameters for the KKRimp selfconsistency workflow.'
        )
        spec.input(
            'voro_params_overwrite',
            valid_type=Dict,
            required=False,
            help=
            'If given, overwrite the some parameters used as input for auxiliary voronoi calculation of starting potential.'
        )
        spec.input(
            'params_kkr_overwrite',
            valid_type=Dict,
            required=False,
            help='Set some input parameters of the KKR calculation for the GF writeout step.'
        )
        spec.input(
            'startpot',
            valid_type=SinglefileData,
            required=False,
            help='Set starting potential (e.g. from preconverged calculation'
        )
        spec.expose_inputs(kkr_flex_wc, namespace='gf_writeout', include=('params_kkr_overwrite', 'options', 'kkr'))

        # structure of the workflow
        spec.outline(
            cls.start,                                                          # initialize workflow
            if_(cls.validate_input)(                                            # validate the input (if true, run_gf_writeout, else skip)
                cls.run_gf_writeout),                                           # write out the host GF
            if_(cls.has_starting_potential_input)(                              # check if strarting potential exists in input already (otherwise create it)
                cls.run_voroaux,                                                  # calculate the auxiliary impurity potentials
                cls.construct_startpot),                                          # construct the host-impurity startpotential
            cls.run_kkrimp_scf,                                                 # run the kkrimp_sub workflow to converge the host-imp startpot
            cls.return_results,                                                 # check if the calculation was successful and return the result nodes
            cls.error_handler)

        # define the possible exit codes
        spec.exit_code(
            141,
            'ERROR_INVALID_INPUT_CODE',
            message='ERROR: one or more of the codes you provided do not '
            'use the necessary plugins: kkr.voro, kkr.kkr, kkr.kkrimp'
        )
        spec.exit_code(
            142,
            'ERROR_MISSING_KKRCODE',
            message='ERROR: since GF writeout step has to be conducted, '
            "'kkrcode' is needed as an input"
        )
        spec.exit_code(
            143,
            'ERROR_MISSING_REMOTE',
            message='ERROR: neither converged host remote nor GF writeout '
            'remote is given as an input. One of them is needed to '
            'proceed with this workflow!'
        )
        spec.exit_code(
            144, 'ERROR_KKRIMP_SUB_WORKFLOW_FAILURE', message='ERROR: sub-workflow for KKRimp convergence failed'
        )
        spec.exit_code(
            145,
            'ERROR_KKRSTARTPOT_WORKFLOW_FAILURE',
            message='ERROR: sub-workflow Kkr_startpot failed (look for failure of voronoi calculation).'
        )

        # define the outputs of the workflow
        spec.output('workflow_info', valid_type=Dict)
        spec.output('last_calc_output_parameters', valid_type=Dict)
        spec.output('last_calc_info', valid_type=Dict)
        spec.output('converged_potential', valid_type=SinglefileData, required=False)
        spec.output('remote_data_gf', valid_type=RemoteData)

    def start(self):
        """
        Init context and some parameters
        """

        self.report(f'INFO: started KKR impurity workflow version {self._workflowversion}')

        # get input parameters
        if 'options' in self.inputs:
            options_dict = self.inputs.options.get_dict()
        else:
            options_dict = self._options_default
            self.report('INFO: using default options')
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = self._wf_default
            self.report('INFO: using default workflow parameters for KKRimp scf cycle')
        if 'voro_aux_parameters' in self.inputs:
            voro_aux_dict = self.inputs.voro_aux_parameters.get_dict()
        else:
            voro_aux_dict = self._voro_aux_default
            self.report('INFO: using default workflow parameters for auxiliary voronoi calculation')

        # get parameters that will be added/overwritten for voronoi run
        if 'voro_params_overwrite' in self.inputs:
            self.ctx.change_voro_params = self.inputs.voro_params_overwrite.get_dict()
        else:
            self.ctx.change_voro_params = {}

        # set option parameters from input, or defaults
        self.ctx.exit_code = None  # collect errors here which are passed on to the error handler in the end
        self.ctx.withmpi = options_dict.get('withmpi', self._options_default['withmpi'])
        self.ctx.resources = options_dict.get('resources', self._options_default['resources'])
        self.ctx.max_wallclock_seconds = options_dict.get(
            'max_wallclock_seconds', self._options_default['max_wallclock_seconds']
        )
        self.ctx.queue = options_dict.get('queue_name', self._options_default['queue_name'])
        self.ctx.custom_scheduler_commands = options_dict.get(
            'custom_scheduler_commands', self._options_default['custom_scheduler_commands']
        )
        self.ctx.options_params_dict = Dict({
            'withmpi': self.ctx.withmpi,
            'resources': self.ctx.resources,
            'max_wallclock_seconds': self.ctx.max_wallclock_seconds,
            'queue_name': self.ctx.queue,
            'custom_scheduler_commands': self.ctx.custom_scheduler_commands
        })
        if 'options_voronoi' in self.inputs:
            self.ctx.options_params_dict_voronoi = self.inputs.options_voronoi.get_dict()
            self.report(f'INFO: Use different options for voronoi code ({self.ctx.options_params_dict_voronoi})')
        else:
            self.ctx.options_params_dict_voronoi = self.ctx.options_params_dict.get_dict()

        # set label and description of the workflow
        self.ctx.description_wf = self.inputs.get(
            'description', 'Workflow for a KKR impurity calculation starting from a host-impurity potential'
        )
        self.ctx.label_wf = self.inputs.get('label', 'kkr_imp_sub_wc')

        # set parameters for the auxiliary voronoi calculation
        self.ctx.voro_dos_params = voro_aux_dict.get('dos_params', self._voro_aux_default['dos_params'])
        self.ctx.voro_num_rerun = voro_aux_dict.get('num_rerun', self._voro_aux_default['num_rerun'])
        self.ctx.voro_fac_cls_increase = voro_aux_dict.get(
            'fac_cls_increase', self._voro_aux_default['fac_cls_increase']
        )
        self.ctx.voro_natom_in_cls_min = voro_aux_dict.get(
            'natom_in_cls_min', self._voro_aux_default['natom_in_cls_min']
        )
        self.ctx.voro_delta_e_min = voro_aux_dict.get('delta_e_min', self._voro_aux_default['delta_e_min'])
        self.ctx.voro_threshold_dos_zero = voro_aux_dict.get(
            'threshold_dos_zero', self._voro_aux_default['threshold_dos_zero']
        )
        self.ctx.voro_check_dos = voro_aux_dict.get('check_dos', self._voro_aux_default['check_dos'])
        #self.ctx.voro_delta_e_min_core_states = voro_aux_dict.get(
        #    'delta_e_min_core_states', self._voro_aux_default['delta_e_min_core_states']
        #)
        # set up new parameter dict to pass to voronoi subworkflow later
        self.ctx.voro_params_dict = Dict({
            'queue_name': self.ctx.queue,
            'resources': self.ctx.resources,
            'max_wallclock_seconds': self.ctx.max_wallclock_seconds,
            'withmpi': self.ctx.withmpi,
            'custom_scheduler_commands': self.ctx.custom_scheduler_commands,
            'dos_params': self.ctx.voro_dos_params,
            'num_rerun': self.ctx.voro_num_rerun,
            'fac_cls_increase': self.ctx.voro_fac_cls_increase,
            'natom_in_cls_min': self.ctx.voro_natom_in_cls_min,
            'delta_e_min': self.ctx.voro_delta_e_min,
            'threshold_dos_zero': self.ctx.voro_threshold_dos_zero,
            'check_dos': self.ctx.voro_check_dos,
        })

        # set workflow parameters for the KKR impurity calculation
        self.ctx.kkr_runmax = wf_dict.get('kkr_runmax', self._wf_default['kkr_runmax'])
        self.ctx.convergence_criterion = wf_dict.get('convergence_criterion', self._wf_default['convergence_criterion'])
        self.ctx.mixreduce = wf_dict.get('mixreduce', self._wf_default['mixreduce'])
        self.ctx.threshold_aggressive_mixing = wf_dict.get(
            'threshold_aggressive_mixing', self._wf_default['threshold_aggressive_mixing']
        )
        self.ctx.strmix = wf_dict.get('strmix', self._wf_default['strmix'])
        self.ctx.nsteps = wf_dict.get('nsteps', self._wf_default['nsteps'])
        self.ctx.aggressive_mix = wf_dict.get('aggressive_mix', self._wf_default['aggressive_mix'])
        self.ctx.aggrmix = wf_dict.get('aggrmix', self._wf_default['aggrmix'])
        self.ctx.broyden_number = wf_dict.get('broyden-number', self._wf_default['broyden-number'])
        self.ctx.mag_init = wf_dict.get('mag_init', self._wf_default['mag_init'])
        self.ctx.hfield = wf_dict.get('hfield', self._wf_default['hfield'])
        self.ctx.init_pos = wf_dict.get('init_pos', self._wf_default['init_pos'])
        self.ctx.accuracy_params = wf_dict.get('accuracy_params', self._wf_default['accuracy_params'])
        # set up new parameter dict to pass to kkrimp subworkflow later
        self.ctx.kkrimp_params_dict = Dict({
            'nsteps': self.ctx.nsteps,
            'kkr_runmax': self.ctx.kkr_runmax,
            'threshold_aggressive_mixing': self.ctx.threshold_aggressive_mixing,
            'convergence_criterion': self.ctx.convergence_criterion,
            'mixreduce': self.ctx.mixreduce,
            'strmix': self.ctx.strmix,
            'aggressive_mix': self.ctx.aggressive_mix,
            'aggrmix': self.ctx.aggrmix,
            'broyden-number': self.ctx.broyden_number,
            'mag_init': self.ctx.mag_init,
            'hfield': self.ctx.hfield,
            'init_pos': self.ctx.init_pos,
            'accuracy_params': self.ctx.accuracy_params,
        })

        # retrieve option for kkrlfex files
        self.ctx.retrieve_kkrflex = wf_dict.get('retrieve_kkrflex', self._wf_default['retrieve_kkrflex'])

        # report the chosen parameters to the user
        self.report(
            'INFO: use the following parameter:\n'
            '\nGeneral settings\n'
            'use mpi: {}\n'
            'resources: {}\n'
            'walltime (s): {}\n'
            'queue name: {}\n'
            'scheduler command: {}\n'
            'description: {}\n'
            'label: {}\n'
            'parameters for the voroaux calculation: {}\n'
            'parameters for the kkrimp scf: {}\n'
            ''.format(
                self.ctx.withmpi, self.ctx.resources, self.ctx.max_wallclock_seconds, self.ctx.queue,
                self.ctx.custom_scheduler_commands, self.ctx.description_wf, self.ctx.label_wf,
                self.ctx.voro_params_dict.get_dict(), self.ctx.kkrimp_params_dict.get_dict()
            )
        )

    def validate_input(self):
        """
        Validate the input and catch possible errors from the input
        """

        inputs = self.inputs
        inputs_ok = True

        if 'kkrimp' and 'voronoi' in inputs:
            try:
                test_and_get_codenode(inputs.kkrimp, 'kkr.kkrimp', use_exceptions=True)
                test_and_get_codenode(inputs.voronoi, 'kkr.voro', use_exceptions=True)
            except ValueError:
                inputs_ok = False
                self.report(self.exit_codes.ERROR_INVALID_INPUT_CODE)  # pylint: disable=no-member
                self.ctx.exit_code = self.exit_codes.ERROR_INVALID_INPUT_CODE  # pylint: disable=no-member
                return False
        elif 'kkr' in inputs:
            try:
                test_and_get_codenode(inputs.kkr, 'kkr.kkr', use_exceptions=True)
            except ValueError:
                inputs_ok = False
                self.report(self.exit_codes.ERROR_INVALID_INPUT_CODE)  # pylint: disable=no-member
                self.ctx.exit_code = self.exit_codes.ERROR_INVALID_INPUT_CODE  # pylint: disable=no-member
                return False

        if 'impurity_info' in inputs:
            self.report(f'INFO: found the following impurity info node in input: {inputs.impurity_info.get_dict()}')

        if 'remote_data_gf' in inputs and 'remote_data_host' in inputs:
            self.report(
                'INFO: both converged host remote (pid: {}) and GF writeout remote (pid: {}) found in input. '
                'Converged host remote will not be used. Skip GF writeout step and '
                'start workflow with auxiliary voronoi calculations.'.format(
                    inputs.remote_data_host.pk, inputs.remote_data_gf.pk
                )
            )
            do_gf_calc = False
        elif 'remote_data_host' in inputs:
            self.report(
                'INFO: found converged host remote (pid: {}) in input. '
                'Start workflow by calculating the host GF.'.format(inputs.remote_data_host.pk)
            )
            if 'kkr' in inputs:
                do_gf_calc = True
            else:
                inputs_ok = False
                self.report(self.exit_codes.ERROR_MISSING_KKRCODE)  # pylint: disable=no-member
                self.ctx.exit_code = self.exit_codes.ERROR_MISSING_KKRCODE  # pylint: disable=no-member
                return False
        elif 'remote_data_gf' in inputs:
            remote_data_gf_node = load_node(inputs.remote_data_gf.pk)
            pk_kkrflex_writeoutcalc = remote_data_gf_node.get_incoming(link_label_filter=u'remote_folder'
                                                                       ).first().node.pk
            self.report(
                'INFO: found remote_data node (pid: {}) from previous KKRFLEX calculation (pid: {}) in input. '
                'Skip GF writeout step and start workflow by auxiliary voronoi calculations.'.format(
                    inputs.remote_data_gf.pk, pk_kkrflex_writeoutcalc
                )
            )
            do_gf_calc = False
            # check if second remote_data_gf node is given (used to overwrite Fermi level)
            if 'remote_data_gf_Efshift' in inputs:
                self.report(
                    'INFO: found remote_data_gf_Efshift (pid: {}) used to overwrite Fermi level.'.format(
                        inputs.remote_data_gf_Efshift.pk
                    )
                )
        else:
            inputs_ok = False
            self.report(self.exit_codes.ERROR_MISSING_REMOTE)  # pylint: disable=no-member
            self.ctx.exit_code = self.exit_codes.ERROR_MISSING_REMOTE  # pylint: disable=no-member
            return False

        self.ctx.do_gf_calc = do_gf_calc
        self.report(f'INFO: validated input successfully: {inputs_ok}. Do GF writeout calc: {self.ctx.do_gf_calc}.')

        return do_gf_calc

    def run_gf_writeout(self):
        """
        Run the gf_writeout workflow to calculate the host Green's function and the
        KKR flexfiles using the converged host remote folder and the impurity info node
        """

        # collect inputs
        kkrcode = self.inputs.kkr
        imp_info = self.inputs.impurity_info
        converged_host_remote = self.inputs.remote_data_host
        options = self.ctx.options_params_dict

        # set label and description of the calc
        sub_label = f'GF writeout (conv. host pid: {converged_host_remote.pk}, imp_info pid: {imp_info.pk})'
        sub_description = 'GF writeout sub workflow for kkrimp_wc using converged host remote data (pid: {}) and impurity_info node (pid: {})'.format(
            converged_host_remote.pk, imp_info.pk
        )

        builder = kkr_flex_wc.get_builder()

        builder.metadata.label = sub_label  # pylint: disable=no-member
        builder.metadata.description = sub_description  # pylint: disable=no-member
        builder.kkr = kkrcode
        builder.options = options
        builder.remote_data = converged_host_remote
        builder.impurity_info = imp_info

        if 'params_kkr_overwrite' in self.inputs:
            builder.params_kkr_overwrite = self.inputs.params_kkr_overwrite

        if 'gf_writeout' in self.inputs:
            if 'options' in self.inputs.gf_writeout:
                builder.options = self.inputs.gf_writeout.options
            if 'params_kkr_overwrite' in self.inputs.gf_writeout:
                builder.params_kkr_overwrite = self.inputs.gf_writeout.params_kkr_overwrite

        # maybe set kkrflex_retrieve
        wf_params_gf = {}
        if not self.ctx.retrieve_kkrflex:
            wf_params_gf['retrieve_kkrflex'] = self.ctx.retrieve_kkrflex
        wf_params_gf = Dict(wf_params_gf)
        builder.wf_parameters = wf_params_gf

        future = self.submit(builder)

        self.report(f'INFO: running GF writeout (pk: {future.pk})')

        return ToContext(gf_writeout=future, last_calc_gf=future)

    def has_starting_potential_input(self):
        """
        check whether or not a starting potential needs to be created
        """
        # initialize
        self.ctx.create_startpot = True

        # check if startpot exists in input
        # TODO maybe implement some consistency checks
        if 'startpot' in self.inputs:
            self.ctx.startpot_kkrimp = self.inputs.startpot
            self.ctx.create_startpot = False

        if self.ctx.exit_code is not None:
            # skip creation of starting potential by overwriting with True return value
            return True

        return self.ctx.create_startpot

    def run_voroaux(self):
        """
        Perform a voronoi calculation for every impurity charge using the structure
        from the converged KKR host calculation
        """
        # TODO: generalize to multiple impurities

        # collect inputs
        vorocode = self.inputs.voronoi
        kkrcode = self.inputs.kkr
        imp_info = self.inputs.impurity_info
        voro_params = self.ctx.voro_params_dict

        if self.ctx.do_gf_calc:
            self.report('INFO: get converged host remote from inputs to extract structure for Voronoi calculation')
            converged_host_remote = self.inputs.remote_data_host
        else:
            self.report(
                'INFO: get converged host remote from GF_host_calc and graph to extract structure for Voronoi calculation'
            )
            remote_data_gf_node = load_node(self.inputs.remote_data_gf.pk)
            GF_host_calc = remote_data_gf_node.get_incoming(link_label_filter=u'remote_folder').first().node
            converged_host_remote = GF_host_calc.inputs.parent_folder

        # find host structure
        structure_host, voro_calc = VoronoiCalculation.find_parent_structure(converged_host_remote)

        # get previous kkr parameters following remote_folder->calc->parameters links
        prev_kkrparams = converged_host_remote.get_incoming(link_label_filter='remote_folder'
                                                            ).first().node.get_incoming(link_label_filter='parameters'
                                                                                        ).first().node
        calc_params = prev_kkrparams

        # set Fermi level for auxiliary impurity potential correctly (extract from EF that is used in impurity calc)
        set_efermi = self.get_ef_from_parent()
        self.report(f'INFO: set Fermi level in jellium starting potential to {set_efermi}')
        # change voronoi parameters
        updatenode = Dict({'ef_set': set_efermi, 'add_direct': True})
        updatenode.label = 'Added Fermi energy'
        voro_params = update_params_wf(voro_params, updatenode)

        # add or overwrite some parameters (e.g. things that are only used by voronoi)
        calc_params_dict = calc_params.get_dict()
        # add some voronoi specific parameters automatically if found (RMTREF should also set RMTCORE to the same value)
        if calc_params_dict.get('<RMTREF>', None) is not None:
            self.report('INFO: add rmtcore to voro params')
            self.ctx.change_voro_params['<RMTCORE>'] = calc_params_dict['<RMTREF>']
            self.report(self.ctx.change_voro_params)

        # add some voronoi-specific settings starting from host's voronoi run (might have been overwritten in between)
        # this is necessary to make sure that voronoi creates the same radial mesh for the impurity potential, otherwise KKRimp will fail
        if voro_calc.inputs.parameters.get_dict().get('RUNOPT', None) is not None:
            self.report("INFO: copy runopt from host's voronoi run")
            runopt = self.ctx.change_voro_params.get('RUNOPT', None)
            if runopt is None:
                runopt = []
            runopt += voro_calc.inputs.parameters['RUNOPT']
            self.ctx.change_voro_params['RUNOPT'] = runopt
        if voro_calc.inputs.parameters.get_dict().get('RCLUSTZ', None) is not None:
            self.report("INFO: copy RCLUSTZ from host's voronoi run")
            self.ctx.change_voro_params['RCLUSTZ'] = voro_calc.inputs.parameters['RCLUSTZ']

        changed_params = False
        for key, val in self.ctx.change_voro_params.items():
            if key in ['RUNOPT', 'TESTOPT']:
                opt_old = calc_params_dict.get(key, [])
                if type(val) != list:
                    val = [val]
                val = opt_old + val
            calc_params_dict[key] = val
            changed_params = True
        if changed_params:
            updatenode = Dict(calc_params_dict)
            updatenode.label = f'Changed params for voroaux: {list(self.ctx.change_voro_params.keys())}'
            updatenode.description = 'Overwritten voronoi input parameter from kkr_imp_wc input.'
            calc_params = update_params_wf(calc_params, updatenode)

        # for every impurity, generate a structure and launch the voronoi workflow
        # to get the auxiliary impurity startpotentials
        self.ctx.voro_calcs = {}
        inter_struc = change_struc_imp_aux_wf(structure_host, imp_info)
        sub_label = f"voroaux calc for Zimp: {imp_info.get_dict().get('Zimp')} in host-struc"
        sub_description = 'Auxiliary voronoi calculation for an impurity with charge '
        sub_description += '{} in the host structure from pid: {}'.format(
            imp_info.get_dict().get('Zimp'), converged_host_remote.pk
        )

        builder = kkr_startpot_wc.get_builder()
        builder.metadata.label = sub_label  # pylint: disable=no-member
        builder.metadata.description = sub_description  # pylint: disable=no-member
        builder.structure = inter_struc
        builder.voronoi = vorocode
        builder.kkr = kkrcode
        builder.wf_parameters = voro_params
        builder.calc_parameters = calc_params
        builder.options = Dict(self.ctx.options_params_dict_voronoi)
        future = self.submit(builder)

        tmp_calcname = f'voro_aux_{1}'
        self.ctx.voro_calcs[tmp_calcname] = future
        self.report(f"INFO: running voro aux (Zimp= {imp_info.get_dict().get('Zimp')}, pid: {future.pk})")

        return ToContext(last_voro_calc=future)

    def get_ef_from_parent(self):
        """
        Extract Fermi level in Ry to which starting potential is set
        """
        # first choose calculation to start from (3 possibilities)
        if self.ctx.do_gf_calc:
            parent_remote = self.inputs.remote_data_host
        elif 'remote_data_gf_Efshift' in self.inputs:
            parent_remote = load_node(self.inputs.remote_data_gf_Efshift.pk)
        else:
            parent_remote = load_node(self.inputs.remote_data_gf.pk)

        # now extract output parameters
        parent_calc = parent_remote.get_incoming(link_label_filter='remote_folder').first().node
        output_params = parent_calc.outputs.output_parameters.get_dict()

        # get fermi energy in Ry from output of KkrCalculation and return result
        set_efermi = output_params.get('fermi_energy')

        return set_efermi

    def construct_startpot(self):
        """
        Take the output of GF writeout and the converged host potential as well as the
        auxiliary startpotentials for the impurity to construct the startpotential for the
        KKR impurity sub workflow
        """

        if not self.ctx.last_voro_calc.is_finished_ok:
            self.report(self.exit_codes.ERROR_KKRSTARTPOT_WORKFLOW_FAILURE)  # pylint: disable=no-member
            self.ctx.exit_code = self.exit_codes.ERROR_KKRSTARTPOT_WORKFLOW_FAILURE  # pylint: disable=no-member

        # collect all nodes necessary to construct the startpotential
        if self.ctx.do_gf_calc:
            GF_host_calc_pk = self.ctx.gf_writeout.outputs.workflow_info.get_dict().get('pk_flexcalc')
            self.report(f'GF_host_calc_pk: {GF_host_calc_pk}')
            GF_host_calc = load_node(GF_host_calc_pk)
            converged_host_remote = self.inputs.remote_data_host
        else:
            remote_data_gf_node = load_node(self.inputs.remote_data_gf.pk)
            GF_host_calc = remote_data_gf_node.get_incoming(link_label_filter=u'remote_folder').first().node
            self.report(f'GF_host_calc_pk: {GF_host_calc.pk}')
            # follow parent_folder link up to get remote folder
            converged_host_remote = GF_host_calc.get_incoming(link_label_filter='parent_folder').first().node

        # get remote folder of last voronoi calculation (i.e. the one from where we take the starting potential)
        print(self.ctx.last_voro_calc)
        all_nodes = self.ctx.last_voro_calc.get_outgoing(node_class=CalcJobNode).all()
        print(all_nodes)
        pk_last_voronoi = max([i.node.pk for i in all_nodes])
        print(pk_last_voronoi)
        voro_calc_remote = load_node(pk_last_voronoi).outputs.remote_folder
        print(voro_calc_remote)

        print(load_node(pk_last_voronoi).outputs.retrieved.list_object_names())
        print(GF_host_calc)
        print(GF_host_calc.outputs.retrieved.list_object_names())

        # check wether or not calculation was taked from cached node
        caching_info = f'INFO: cache_source of GF_host_calc node: {GF_host_calc.get_cache_source()}'
        print(caching_info)
        self.report(caching_info)
        caching_info = f'INFO: cache_source of voronoi node: {load_node(pk_last_voronoi).get_cache_source()}'
        print(caching_info)
        self.report(caching_info)

        imp_info = self.inputs.impurity_info
        nspin = GF_host_calc.outputs.output_parameters.get_dict().get('nspin')

        ilayer_cent = imp_info.get_dict().get('ilayer_center', 0)  # defaults to layer 0

        # prepare settings dict
        potname_converged = 'out_potential'
        potname_impvorostart = 'output.pot'
        potname_imp = 'potential_imp'

        if nspin < 2:
            replacelist_pot2 = [[0, ilayer_cent]]
        else:
            replacelist_pot2 = [[0, 2 * ilayer_cent], [1, 2 * ilayer_cent + 1]]
        try:
            with GF_host_calc.outputs.retrieved.open('scoef') as _f:
                neworder_pot1 = [int(i) for i in np.loadtxt(_f, skiprows=1)[:, 3] - 1]
        except:
            with GF_host_calc.outputs.retrieved.open('scoef') as _f:
                neworder_pot1 = [int(np.loadtxt(_f, skiprows=1)[3] - 1)]

        settings_label = f'startpot_KKRimp for imp_info node {imp_info.pk}'
        settings_description = f'starting potential for impurity info: {imp_info}'

        settings = Dict({
            'out_pot': potname_imp,
            'neworder': neworder_pot1,
            'replace_newpos': replacelist_pot2,
            'label': settings_label,
            'description': settings_description
        })
        print('startpot_kkrimp construction:', settings, converged_host_remote, voro_calc_remote)
        startpot_kkrimp = neworder_potential_wf(
            settings_node=settings, parent_calc_folder=converged_host_remote, parent_calc_folder2=voro_calc_remote
        )

        # add starting potential for kkrimp calculation to context
        self.ctx.startpot_kkrimp = startpot_kkrimp

        self.report(
            'INFO: created startpotential (pid: {}) for the impurity calculation '
            'by using information of the GF host calculation (pid: {}), the potential of the '
            'converged host system (remote pid: {}) and the potential of the auxiliary voronoi '
            'calculation (remote pid: {})'.format(
                startpot_kkrimp.pk, GF_host_calc.pk, converged_host_remote.pk, self.ctx.last_voro_calc.pk
            )
        )

    def run_kkrimp_scf(self):
        """
        Uses both the previously generated host-impurity startpotential and the output from
        the GF writeout workflow as inputs to run the kkrimp_sub workflow in order to
        converge the host-impurity potential
        """

        # collect all necessary input nodes
        kkrimpcode = self.inputs.kkrimp
        startpot = self.ctx.startpot_kkrimp
        kkrimp_params = self.ctx.kkrimp_params_dict
        options = self.ctx.options_params_dict
        imp_info = self.inputs.impurity_info
        if self.ctx.do_gf_calc:
            self.report(f'INFO: get GF remote from gf_writeout sub wf (pid: {self.ctx.gf_writeout.pk})')
            gf_remote = self.ctx.gf_writeout.outputs.GF_host_remote
        else:
            self.report(f'INFO: get GF remote from input node (pid: {self.inputs.remote_data_gf.pk})')
            gf_remote = self.inputs.remote_data_gf
        # save in context to return as output node
        self.ctx.gf_remote = gf_remote

        # set label and description
        sub_label = f'kkrimp_sub scf wf (GF host remote: {gf_remote.pk}, imp_info: {self.inputs.impurity_info.pk})'
        sub_description = 'convergence of the host-impurity potential (pk: {}) using GF remote (pk: {})'.format(
            startpot.pk, gf_remote.pk
        )

        builder = kkr_imp_sub_wc.get_builder()
        builder.metadata.label = sub_label  # pylint: disable=no-member
        builder.metadata.description = sub_description  # pylint: disable=no-member
        builder.kkrimp = kkrimpcode
        builder.options = options
        builder.impurity_info = imp_info
        builder.host_imp_startpot = startpot
        builder.remote_data = gf_remote
        if 'remote_data_gf_Efshift' in self.inputs:
            builder.remote_data_Efshift = self.inputs.remote_data_gf_Efshift
        if 'scf' in self.inputs:
            if 'params_overwrite' in self.inputs.scf:
                builder.params_overwrite = self.inputs.scf.params_overwrite
            if 'options' in self.inputs.scf:
                builder.options = self.inputs.scf.options
            if 'initial_noco_angles' in self.inputs.scf:
                builder.initial_noco_angles = self.inputs.scf.initial_noco_angles
            if 'rimpshift' in self.inputs.scf:
                builder.rimpshift = self.inputs.scf.rimpshift
            if 'settings_LDAU' in self.inputs.scf:
                builder.settings_LDAU = self.inputs.scf.settings_LDAU
        builder.wf_parameters = kkrimp_params
        future = self.submit(builder)

        self.report(
            f'INFO: running kkrimp_sub_wf (startpot: {startpot.pk}, GF_remote: {gf_remote.pk}, wf pid: {future.pk})'
        )

        return ToContext(kkrimp_scf_sub=future)

    def return_results(self):
        """
        Return the results and create all of the output nodes
        """

        self.report('INFO: creating output nodes for the KKR impurity workflow ...')

        if self.ctx.kkrimp_scf_sub.is_finished_ok:
            link_nodes = {'results_scf_workflow': self.ctx.kkrimp_scf_sub.outputs.workflow_info}

            last_calc_pk = self.ctx.kkrimp_scf_sub.outputs.workflow_info.get_dict().get('last_calc_nodeinfo')['pk']
            last_calc_output_params = load_node(last_calc_pk).outputs.output_parameters
            last_calc_info = self.ctx.kkrimp_scf_sub.outputs.workflow_info
            outputnode_dict = {}
            outputnode_dict['workflow_name'] = self.__class__.__name__
            outputnode_dict['workflow_version'] = self._workflowversion
            if self.ctx.do_gf_calc:
                outputnode_dict['used_subworkflows'] = {
                    'gf_writeout': self.ctx.gf_writeout.pk,
                    'kkr_imp_sub': self.ctx.kkrimp_scf_sub.pk
                }
                outputnode_dict['gf_wc_success'] = self.ctx.gf_writeout.outputs.workflow_info.get_dict(
                ).get('successful')
                link_nodes['gf_writeout'] = self.ctx.gf_writeout.outputs.workflow_info
            else:
                outputnode_dict['used_subworkflows'] = {'kkr_imp_sub': self.ctx.kkrimp_scf_sub.pk}
            if self.ctx.create_startpot:
                outputnode_dict['used_subworkflows']['auxiliary_voronoi'] = self.ctx.last_voro_calc.pk
                res_voro_info = self.ctx.last_voro_calc.outputs.results_vorostart_wc
                outputnode_dict['voro_wc_success'] = res_voro_info.get_dict().get('successful')
                link_nodes['results_startpot_workflow'] = self.ctx.last_voro_calc.outputs.results_vorostart_wc
            outputnode_dict['converged'] = last_calc_info.get_dict().get('convergence_reached')
            outputnode_dict['number_of_rms_steps'] = len(last_calc_info.get_dict().get('convergence_values_all_steps'))
            outputnode_dict['convergence_values_all_steps'] = last_calc_info.get_dict(
            ).get('convergence_values_all_steps')
            outputnode_dict['impurity_info'] = self.inputs.impurity_info.get_dict()
            outputnode_dict['kkrimp_wc_success'] = last_calc_info.get_dict().get('successful')
            outputnode_dict['last_calculation_uuid'] = load_node(last_calc_pk).uuid

            # create results node and link all sub-workflow output nodes
            outputnode_t = create_out_dict_node(Dict(dict=outputnode_dict), **link_nodes)
            outputnode_t.label = 'kkrimp_wc_inform'
            outputnode_t.description = 'Contains information for workflow'
            self.report(f'INFO: workflow_info node: {outputnode_t.uuid}')

            self.out('workflow_info', outputnode_t)
            self.out('last_calc_output_parameters', last_calc_output_params)
            self.out('last_calc_info', last_calc_info)
            self.out('converged_potential', self.ctx.kkrimp_scf_sub.outputs.host_imp_pot)
            self.out('remote_data_gf', self.ctx.gf_remote)

            # print final message before exiting
            self.report('INFO: created 3 output nodes for the KKR impurity workflow.')
            self.report(
                '\n'
                '|------------------------------------------------------------------------------------------------------------------|\n'
                '|-------------------------------------| Done with the KKR impurity workflow! |-------------------------------------|\n'
                '|------------------------------------------------------------------------------------------------------------------|'
            )
        else:
            self.report(self.exit_codes.ERROR_KKRIMP_SUB_WORKFLOW_FAILURE)  # pylint: disable=no-member
            return self.exit_codes.ERROR_KKRIMP_SUB_WORKFLOW_FAILURE  # pylint: disable=no-member

    def error_handler(self):
        """Capture errors raised in validate_input"""
        if self.ctx.exit_code is not None:
            return self.ctx.exit_code


@calcfunction
def change_struc_imp_aux_wf(struc, imp_info):  # Note: works for single imp at center only!
    from aiida.common.constants import elements as PeriodicTableElements
    _atomic_numbers = {data['symbol']: num for num, data in PeriodicTableElements.items()}

    new_struc = StructureData(cell=struc.cell)
    new_struc.pbc = struc.pbc  # take also pbc values from parent struc
    isite = 0
    for site in struc.sites:
        sname = site.kind_name
        kind = struc.get_kind(sname)
        pos = site.position
        # intermediate fix to avoid crash for old structures with vacuum:'{H0.00X1.00}'
        # use atom kind='X' in the future for new structures
        if kind.get_symbols_string() == '{H0.00X1.00}':
            zatom = 0
        else:
            zatom = _atomic_numbers[kind.get_symbols_string()]
        if isite == imp_info.get_dict().get('ilayer_center', 0):
            zatom = imp_info.get_dict().get('Zimp')
            if type(zatom) == list:
                zatom = zatom[0]  # here this works for single impurity only!
        symbol = PeriodicTableElements.get(zatom).get('symbol')
        new_struc.append_atom(position=pos, symbols=symbol)
        isite += 1

    return new_struc



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-KKR package.                                 #
#                                                                             #
# For further information on the license, see the LICENSE.txt file            #
###############################################################################
# pylint: disable=cyclic-import
"""Command line utilities to create and inspect `Dict` nodes with FLAPW parameters."""
import click
from aiida.cmdline.params import options
from aiida.cmdline.utils import decorators, echo
from . import cmd_data
from aiida.orm import Dict
from aiida_kkr.tools import kkrparams


@cmd_data.group('parameter')
def cmd_parameter():
    """Commands to create and inspect `Dict` nodes containing FLAPW parameters ('calc_parameters')."""


@cmd_parameter.command('import')
@click.argument('filename', type=click.Path(exists=True))
@click.option('--show/--no-show', default=True, show_default=True, help='Print the contents from the extracted dict.')
@click.option('--verbose', '-v', is_flag=True, default=False, show_default=True, help='Print verbose output.')
@options.DRY_RUN()
@decorators.with_dbenv()
def cmd_param_import(filename, show, verbose, dry_run):
    """
    Import kkr params Dict from a KKRhost input file.

    FILENAME is the name/path of the inputcard file to use.
    """

    kkr_para = kkrparams()
    if verbose:
        print(f'start reading from file: {filename}')
    kkr_para.read_keywords_from_inputcard(filename)
    if verbose:
        print(f'read kkr params: {kkr_para.get_set_values()}')
    missing = kkr_para.get_missing_keys(use_aiida=True)
    if len(missing) > 0:
        # TODO replace by click echo
        echo.echo_warning(
            f"""inputcard did not contain all necessary keys to run a kkr calculation!
The following keys are missing:
{missing}"""
        )

    struc_keys = ['BRAVAIS', '<RBASIS>', '<ZATOM>', 'ALATBASIS', 'NAEZ', 'NATYP']
    no_struc_kkrparams = {k: v for k, v in kkr_para.get_set_values() if k not in struc_keys}
    if verbose:
        print(f'keys after taking out the ones used in the structure node:: {no_struc_kkrparams}')

    if len(no_struc_kkrparams.keys()) == 0:
        echo.echo_critical('failed to extract kkr params')
    else:
        param_node = Dict(no_struc_kkrparams)
        if dry_run:
            echo.echo_success('parsed kkr params from input file')
        else:
            param_node.store()
            message = f'stored kkr params Dict node <{param_node.pk}>'
            echo.echo_success(message)

    if show:
        echo.echo_dictionary(param_node.get_dict())



"""
Here we implement a Structure Data type that is specific to the KKR code
It can store empty spheres, i.e can store 'dummy atom' with charge 0
"""
#from aiida.orm.data.structure import *
from aiida.orm import Data, StructureData
from aiida.plugins import DataFactory


class KkrstructureData(StructureData):
    """
    Extention of AiiDA StructureData type that can store empty spheres/dummy atoms.
    it also provides some methods to convert to and from an aiida StructreData node.
    """

    def __init__(self, *args, **kwargs):
        super(KkrstructureData, self).__init__(*args, **kwargs)

    pass



# -*- coding: utf-8 -*-
"""
Parser for the KKR-impurity Code.
The parser should never fail, but it should catch
all errors and warnings and show them to the user.
"""

import tarfile
import os
from aiida import __version__ as aiida_core_version
from aiida.orm import Dict, CalcJobNode
from aiida.common.exceptions import InputValidationError, NotExistent, NotExistentAttributeError
from aiida.parsers.parser import Parser
from aiida_kkr.calculations.kkrimp import KkrimpCalculation
from aiida_kkr.tools.context import open_files_in_context
from masci_tools.io.parsers.kkrparser_functions import check_error_category
from masci_tools.io.parsers.kkrimp_parser_functions import KkrimpParserFunctions
from pprint import pprint
from contextlib import ExitStack

__copyright__ = (u'Copyright (c), 2018, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.6.2'
__contributors__ = (u'Philipp Rüßmann', u'Raffaele Aliberti')


class KkrimpParser(Parser):
    """
    Parser class for parsing output of the KKRimp code..
    """

    def __init__(self, calc):
        """
        Initialize the instance of KkrimpParser
        """

        self._ParserVersion = __version__

        # reuse init of base class
        super(KkrimpParser, self).__init__(calc)

    # pylint: disable=protected-access
    # pylint: disable=unexpected-keyword-arg
    def parse(self, debug=False, ignore_nan=True, doscalc=None, **kwargs):
        """
        Parse output data folder, store results in database.

        :param retrieved: a dictionary of retrieved nodes, where
          the key is the link name
        """
        success = False
        node_list = ()

        # Check that the retrieved folder is there
        try:
            out_folder = self.retrieved
        except NotExistent:
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER

        # check what is inside the folder
        list_of_files = out_folder.list_object_names()

        file_errors = []
        files = {}

        # Get the node of the parent calculation (useful for the imp calculations)
        calc = out_folder.get_incoming(node_class=CalcJobNode).first().node

        # figure out if this is a DOS calculation (if not specified in the input already)
        if doscalc is None:
            parent_host_calc = calc.inputs.host_Greenfunction_folder.get_incoming(node_class=CalcJobNode).first().node
            # This parameter discriminates if it is a DOS calc or not
            npol = parent_host_calc.inputs.parameters.get_dict().get('NPOL', 1)
            doscalc = (npol == 0)

        if debug:
            print('doscalc = ', doscalc)

        # Parse output files of KKRimp calculation

        # first get path to files and catch errors if files are not present
        # append tupels (error_category, error_message) where error_category is
        # 1: critical error, always leads to failing of calculation
        # 2: warning, is inspected and checked for consistency with read-in
        #    out_dict values (e.g. nspin, newsosol, ...)

        # first go through list of critical files that should always be there
        critical_files = [
            ('outfile', KkrimpCalculation._DEFAULT_OUTPUT_FILE),
            ('out_log', KkrimpCalculation._OUTPUT_000),
            ('out_pot', KkrimpCalculation._OUT_POTENTIAL),
            ('out_timing', KkrimpCalculation._OUT_TIMING_000),
            ('out_enersp_at', KkrimpCalculation._OUT_ENERGYSP_PER_ATOM),
            ('out_enertot_at', KkrimpCalculation._OUT_ENERGYTOT_PER_ATOM),
        ]
        for keyname, fname in critical_files:
            self._check_file_existance(files, keyname, fname, 1, file_errors)

        additional_files = [
            ('kkrflex_llyfac', KkrimpCalculation._KKRFLEX_LLYFAC),
            ('kkrflex_angles', KkrimpCalculation._KKRFLEX_ANGLE),
            ('out_spinmoms', KkrimpCalculation._OUT_MAGNETICMOMENTS),
            ('out_orbmoms', KkrimpCalculation._OUT_ORBITALMOMENTS),
        ]
        for keyname, fname in additional_files:
            self._check_file_existance(files, keyname, fname, 2, file_errors)

        if debug:
            pprint(files)

        # now parse file output
        out_dict = {
            'parser_version': self._ParserVersion,
            'calculation_plugin_version': KkrimpCalculation._CALCULATION_PLUGIN_VERSION
        }

        # open all files in context managers
        with ExitStack() as stack:
            # open files only if they exist
            # openend files are added to the context manager stack
            # so that at the end of the parsing all files are closed
            # when the context manager is exited
            fhandles = open_files_in_context(stack, out_folder, *files.values())
            named_file_handles = dict([[key, fhandles[ik]] for ik, key in enumerate(files.keys())])

            # now we can parse the output files
            success, msg_list, out_dict = KkrimpParserFunctions().parse_kkrimp_outputfile(
                out_dict, named_file_handles, debug=debug, ignore_nan=ignore_nan, doscalc=doscalc
            )

        out_dict['parser_errors'] = msg_list
        # add file open errors to parser output of error messages
        for (err_cat, f_err) in file_errors:
            if err_cat == 1:
                msg_list.append(f_err)
            elif check_error_category(err_cat, f_err, out_dict):
                msg_list.append(f_err)
            else:
                if 'parser_warnings' not in list(out_dict.keys()):
                    out_dict['parser_warnings'] = []
                out_dict['parser_warnings'].append(f_err.replace('Error', 'Warning'))
        out_dict['parser_errors'] = msg_list

        # create output node and link
        self.out('output_parameters', Dict(dict=out_dict))

        if not success:
            return self.exit_codes.ERROR_PARSING_KKRIMPCALC

    def _check_file_existance(self, files, keyname, fname, icrit, file_errors):
        """Check if file called `fname` exists and then
        add it to the `files` dict with a given `keyname`.
        The icrit index determines how critical it is if a file
        is not found (1=critical error, 2=only a warning).
        """
        if fname in self.retrieved.list_object_names():
            files[keyname] = fname
        else:
            # file not
            if icrit == 1:
                crit_level = 'Critical file error!'
            elif icrit == 2:
                crit_level = 'Warning!'
            else:
                raise ValueError('icrit should be either 1 or 2')
            file_errors.append((icrit, crit_level + f" File '{fname}' not found."))
            files[keyname] = None



# -*- coding: utf-8 -*-
"""
Input plug-in for a KKRimp calculation.
"""

from aiida.engine import CalcJob
from aiida.orm import CalcJobNode, Dict, RemoteData, SinglefileData, Bool
from aiida.common.utils import classproperty
from aiida.common.exceptions import (InputValidationError, ValidationError, UniquenessError)
from aiida.common.datastructures import (CalcInfo, CodeInfo)
from masci_tools.io.kkr_params import kkrparams
from .voro import VoronoiCalculation
from .kkr import KkrCalculation
from aiida_kkr.tools.tools_kkrimp import modify_potential, make_scoef, write_scoef_full_imp_cls, get_imp_info_from_parent
from aiida_kkr.tools.common_workfunctions import get_username
from aiida_kkr.tools.ldau import get_ldaupot_text
from aiida_kkr.tools.imp_cluster_tools import get_scoef_single_imp
from masci_tools.io.common_functions import search_string, get_ef_from_potfile
import os
import tarfile
import numpy as np

__copyright__ = (u'Copyright (c), 2018, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.10.2'
__contributors__ = (u'Philipp Rüßmann', u'Fabian Bertoldo')

#TODO: implement 'ilayer_center' consistency check


class KkrimpCalculation(CalcJob):
    """
    AiiDA calculation plugin for a KKRimp calculation.
    """

    # calculation plugin version
    _CALCULATION_PLUGIN_VERSION = __version__

    # List of mandatory input files
    _CONFIG = u'config.cfg'
    _POTENTIAL = u'potential'
    _KKRFLEX_GREEN = KkrCalculation._KKRFLEX_GREEN
    _KKRFLEX_TMAT = KkrCalculation._KKRFLEX_TMAT
    _KKRFLEX_ATOMINFO = KkrCalculation._KKRFLEX_ATOMINFO
    _KKRFLEX_INTERCELL_REF = KkrCalculation._KKRFLEX_INTERCELL_REF
    _KKRFLEX_INTERCELL_CMOMS = KkrCalculation._KKRFLEX_INTERCELL_CMOMS

    # List of optional input files (may be mandatory for some setting in inputputcard)
    _SHAPEFUN = u'shapefun'
    _KKRFLEX_ANGLE = u'kkrflex_angle'
    _KKRFLEX_LLYFAC = u'kkrflex_llyfac'
    _KKRFLEX_SOCFAC = u'kkrflex_spinorbitperatom'
    _KKRFLEX_RIMPSHIFT = u'kkrflex_rimpshift'

    # full list of kkrflex files
    _ALL_KKRFLEX_FILES = KkrCalculation._ALL_KKRFLEX_FILES
    _ALL_KKRFLEX_FILES.append(_KKRFLEX_ANGLE)
    _ALL_KKRFLEX_FILES.append(_KKRFLEX_LLYFAC)
    _ALL_KKRFLEX_FILES.append(_KKRFLEX_SOCFAC)

    # List of output files that should always be present (are always retrieved)
    _OUT_POTENTIAL = u'out_potential'
    _OUTPUT_000 = u'out_log.000.txt'
    _OUT_TIMING_000 = u'out_timing.000.txt'
    _OUT_ENERGYSP = u'out_energysp_eV'
    _OUT_ENERGYSP_PER_ATOM = u'out_energysp_per_atom_eV'
    _OUT_ENERGYTOT = u'out_energytotal_eV'
    _OUT_ENERGYTOT_PER_ATOM = u'out_energytotal_per_atom_eV'

    # List of output files that are retrieved if special conditions are fulfilled
    _OUT_JIJMAT = u'out_Jijmatrix'
    _OUT_JIJ_OF_E_BASE = 'out_Jijmatrix_Eres_IE*'
    _OUT_LDOS = u'out_ldos.atom=*'
    _OUT_LDOS_INTERPOL = u'out_ldos.interpol.atom=*'
    _OUT_LMDOS = u'out_lmdos.atom=*'
    _OUT_LMDOS_INTERPOL = u'out_lmdos.interpol.atom=*'
    _OUT_MAGNETICMOMENTS = u'out_magneticmoments'
    _OUT_ORBITALMOMENTS = u'out_orbitalmoments'
    _LDAUPOT = u'ldaupot'

    _TEST_RHO2NS = u'test_rho2ns'

    # template.product entry point defined in setup.json
    _default_parser = u'kkr.kkrimpparser'

    # default names used within aiida (verdi calculation out(in)putcat)
    _OUTPUT_FILE_NAME = u'out_kkrimp'
    _INPUT_FILE_NAME = _CONFIG
    _DEFAULT_OUTPUT_FILE = _OUTPUT_FILE_NAME  # will be shown with outputcat
    _DEFAULT_INPUT_FILE = _INPUT_FILE_NAME  # will be shown with inputcat

    # name of tarfile which is created by parser after successful parsing (to reduce amount of data stored in repo)
    _FILENAME_TAR = 'output_all.tar.gz'
    _DIRNAME_GF_UPLOAD = 'kkrflex_green_upload'

    @classmethod
    def define(cls, spec):
        """
        Init internal parameters at class load time
        """

        # reuse base class function
        super(KkrimpCalculation, cls).define(spec)
        # now define input files and parser
        spec.input('metadata.options.parser_name', valid_type=str, default=cls._default_parser, non_db=True)
        spec.input('metadata.options.input_filename', valid_type=str, default=cls._DEFAULT_INPUT_FILE, non_db=True)
        spec.input('metadata.options.output_filename', valid_type=str, default=cls._DEFAULT_OUTPUT_FILE, non_db=True)
        # define input nodes (optional ones have required=False)
        spec.input(
            'parameters',
            valid_type=Dict,
            required=False,
            help='Use a node that specifies the input parameters (calculation settings).'
        )
        spec.input(
            'host_Greenfunction_folder',
            valid_type=RemoteData,
            required=True,
            help=
            'Use a node that specifies the host KKR calculation contaning the host Green function and tmatrix (KkrCalculation with impurity_info input).'
        )
        spec.input(
            'host_Greenfunction_folder_Efshift',
            valid_type=RemoteData,
            required=False,
            help=
            'Use a node that specifies the host KKR calculation contaning the host Green function and tmatrix with Fermi level shift (used to set Fermi level).'
        )
        spec.input(
            'impurity_potential',
            valid_type=SinglefileData,
            required=False,
            help='Use a node that contains the input potential.'
        )
        spec.input(
            'parent_calc_folder',
            valid_type=RemoteData,
            required=False,
            help='Use a node that specifies a parent KKRimp calculation.'
        )
        spec.input(
            'impurity_info',
            valid_type=Dict,
            required=False,
            help='Use a parameter node that specifies properties for a immpurity calculation.'
        )
        spec.input(
            'settings_LDAU',
            valid_type=Dict,
            required=False,
            help="""
Settings for running a LDA+U calculation. The Dict node should be of the form
    settings_LDAU = Dict({'iatom=0':{
        'L': 3,         # l-block which gets U correction (1: p, 2: d, 3: f-electrons)
        'U': 7.,        # U value in eV
        'J': 0.75,      # J value in eV
        'Eref_EF': 0.,  # reference energy in eV relative to the Fermi energy. This is the energy where the projector wavefunctions are calculated (should be close in energy where the states that are shifted lie (e.g. for Eu use the Fermi energy))
    }})

Note: you can add multiple entries like the one for iatom==0 in this example. The atom index refers to the corresponding atom in the impurity cluster.
"""
        )
        spec.input(
            'initial_noco_angles',
            valid_type=Dict,
            required=False,
            help="""
Initial non-collinear angles for the magnetic moments of the impurities. These values will be written into the `kkrflex_angle` input file of KKRimp.
The Dict node should be of the form
    initial_noco_angles = Dict({
        'theta': [theta_at1, theta_at2, ..., theta_atN], # list theta values in degrees (0..180)
        'phi': [phi_at1, phi_at2, ..., phi_atN],         # list phi values in degrees (0..360)
        'fix_dir': [True, False, ..., True/False],       # list of booleans indicating of the direction of the magentic moment should be fixed or is allowed to be updated (True means keep the direction of the magnetic moment fixed)
    })

Note: The length of the theta, phi and fix_dir lists have to be equal to the number of atoms in the impurity cluster.
"""
        )
        spec.input(
            'rimpshift',
            valid_type=Dict,
            required=False,
            help="""
Shift for atoms in the impurity cluster used in U-transformation.

The Dict node should be of the form
    rimpshift = Dict({'shifts': [[0., 0., 0.], ... ]})
where the shifts are given in atomic units (i.e. the internal KKR units).

Note: The length of the 'shifts' attribute should be an array with three numbers indicating the shift for each atom in the impurity cluster.
"""
        )

        # define outputs
        spec.output('output_parameters', valid_type=Dict, required=True, help='results of the KKRimp calculation')
        spec.default_output_node = 'output_parameters'
        # define exit codes, also used in parser
        spec.exit_code(301, 'ERROR_NO_RETRIEVED_FOLDER', message='Retrieved folder of KKRimp calculation not found.')
        spec.exit_code(302, 'ERROR_PARSING_KKRIMPCALC', message='KKRimp parser returned an error.')
        #TBD

    def prepare_for_submission(self, tempfolder):
        """
        Create input files.

            :param tempfolder: aiida.common.folders.Folder subclass where
                the plugin should put all its files.
            :param inputdict: dictionary of the input nodes as they would
                be returned by get_inputs_dict
        """
        # Check inputdict and extrace nodes
        tmp = self._check_and_extract_input_nodes(tempfolder)
        parameters, code, imp_info = tmp[0], tmp[1], tmp[2]
        kkrflex_file_paths, shapefun_path, shapes = tmp[3], tmp[4], tmp[5]
        host_parent_calc, params_host, impurity_potential = tmp[6], tmp[7], tmp[8]
        parent_calc_folder, structure = tmp[9], tmp[10]

        # Prepare input files for KKRimp calculation
        # 1. fill kkr params for KKRimp, write config file and eventually also kkrflex_llyfac file
        self._extract_and_write_config(
            parent_calc_folder, params_host, parameters, tempfolder,
            kkrflex_file_paths[KkrCalculation._KKRFLEX_ATOMINFO]
        )
        # 2. write shapefun from impurity info and host shapefun and copy imp. potential
        self._get_pot_and_shape(
            imp_info, shapefun_path, shapes, impurity_potential, parent_calc_folder, tempfolder, structure
        )
        # 3. change kkrflex_atominfo to match impurity case
        self._change_atominfo(imp_info, kkrflex_file_paths, tempfolder)

        # prepare copy and retrieve lists
        local_copy_list = []  # potential already in tempfolder which is copied in full
        for filename in list(kkrflex_file_paths.keys()):
            if filename != self._KKRFLEX_ATOMINFO:
                src_path = kkrflex_file_paths[filename]
                local_copy_list.append((src_path.uuid, filename, filename))

        retrieve_list = [
            self._OUTPUT_FILE_NAME, self._CONFIG, self._KKRFLEX_ATOMINFO, self._KKRFLEX_ANGLE, self._KKRFLEX_LLYFAC,
            self._KKRFLEX_SOCFAC, self._OUT_POTENTIAL, self._OUTPUT_000, self._OUT_TIMING_000,
            self._OUT_ENERGYSP_PER_ATOM, self._OUT_ENERGYTOT_PER_ATOM
        ]

        # maybe create kkrflex_rimpshift file
        self._write_kkrflex_rimpshift(tempfolder, parameters)

        # extract run and test options (these change retrieve list in some cases)
        allopts = self.get_run_test_opts(parameters)

        # retrieve l(m)dos files
        retrieve_list = self.add_lmdos_files_to_retrieve(tempfolder, allopts, retrieve_list, kkrflex_file_paths)

        # change retrieve list for Chebychev solver
        retrieve_list = self.adapt_retrieve_tmatnew(tempfolder, allopts, retrieve_list)

        # write ldaupot file and change retrieved list for LDA+U calculation
        # this is triggered with having the settings_LDAU dict node in input.
        retrieve_list = self.init_ldau(tempfolder, retrieve_list, parent_calc_folder)

        # add Jij files to retrieve list if calculation if in Jij mode
        retrieve_list = self.add_jij_files(tempfolder, retrieve_list)

        # change local and remote copy list if GF is found on remote machine
        remote_symlink_list, local_copy_list = self.get_remote_symlink(local_copy_list)

        # Prepare CalcInfo to be returned to aiida (e.g. retreive_list etc.)
        calcinfo = CalcInfo()
        calcinfo.uuid = self.uuid
        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = []
        calcinfo.remote_symlink_list = remote_symlink_list
        calcinfo.retrieve_list = retrieve_list

        codeinfo = CodeInfo()
        codeinfo.cmdline_params = []
        codeinfo.stdout_name = self._OUTPUT_FILE_NAME
        codeinfo.code_uuid = code.uuid
        calcinfo.codes_info = [codeinfo]

        return calcinfo

    #####################################################
    # helper functions

    def _get_and_verify_hostfiles(self, tempfolder):
        """
        Check inputdict for host_Greenfunction_folder and extract impurity_info, paths to kkrflex-files and path of shapefun file

        :param inputdict: input dictionary containing all input nodes to KkrimpCalculation
        :returns:
            * imp_info: Dict node containing impurity information like position, Z_imp, cluster size, etc.
            * kkrflex_file_paths: dict of absolute file paths for the kkrflex files
            * shapefun_path: absolute path of the shapefunction file in the host calculation (needed to construct shapefun_imp)
            * shapes: mapping array of atoms to shapes (<SHAPE> input)
        :note: shapefun_path is None if host_Greenfunction calculation was not full-potential
        :raises:
            * InputValidationError, if inputdict does not contain 'host_Greenfunction'
            * InputValidationError, if host_Greenfunction_folder not of right type
            * UniquenessError, if host_Greenfunction_folder does not have exactly one parent
            * InputValidationError, if host_Greenfunction does not have an input node impurity_info
            * InputValidationError, if host_Greenfunction was not a KKRFLEX calculation
        """

        # get mandatory input nodes (extract host_Greenfunction_folder)
        host_parent = self.inputs.host_Greenfunction_folder

        # extract parent calculation
        parent_calcs = host_parent.get_incoming(node_class=CalcJob)
        n_parents = len(parent_calcs.all_link_labels())
        if n_parents != 1:
            raise UniquenessError(
                'Input RemoteData is child of {} '
                'calculation{}, while it should have a single parent'
                ''.format(n_parents, '' if n_parents == 0 else 's')
            )
        else:
            parent_calc = parent_calcs.first().node
        # extract impurity_info
        found_impurity_inputnode = False
        found_host_parent = False
        if 'impurity_info' in self.inputs:
            imp_info_inputnode = self.inputs.impurity_info
            if not isinstance(imp_info_inputnode, Dict):
                raise InputValidationError('impurity_info not of type Dict')
            found_impurity_inputnode = True
            found_host_parent = True
        imp_info = get_imp_info_from_parent(parent_calc)
        if imp_info is None:
            raise InputValidationError('host_Greenfunction calculation does not have an input node impurity_info')

        # if impurity input is seperate input, check if it is the same as
        # the one from the parent calc (except for 'Zimp'). If that's not the
        # case, raise an error
        if found_impurity_inputnode and found_host_parent:
            check_consistency_imp_info = False
            if 'imp_cls' in imp_info_inputnode.get_dict().keys():
                input_imp_cls_arr = np.array(imp_info_inputnode.get_dict()['imp_cls'])

                try:
                    parent_imp_cls_arr = np.array(imp_info.get_dict()['imp_cls'])
                except:
                    host_structure, _ = VoronoiCalculation.find_parent_structure(parent_calc)
                    parent_imp_cls_arr = get_scoef_single_imp(host_structure, imp_info)

                is_identical = np.array_equal(
                    np.round(input_imp_cls_arr[:, 0:4], 5), np.round(parent_imp_cls_arr[:, 0:4], 5)
                )

                if is_identical:
                    check_consistency_imp_info = True
                else:
                    self.report('impurity_info node from input and from previous GF calculation are NOT compatible!.')
                    self.report(f'impurity_info node from input: {input_imp_cls_arr[:, 0:4]}')
                    self.report(f'impurity_info node from previous GF calculation: {parent_imp_cls_arr[:, 0:4]}')

            #TODO: implement also 'ilayer_center' check
            elif imp_info_inputnode.get_dict().get('Rcut') == imp_info.get_dict().get('Rcut'):

                check_consistency_imp_info = True
                try:
                    if (
                        imp_info_inputnode.get_dict().get('hcut') == imp_info.get_dict().get('hcut') and
                        imp_info_inputnode.get_dict().get('cylinder_orient')
                        == imp_info.get_dict().get('cylinder_orient') and
                        imp_info_inputnode.get_dict().get('Rimp_rel') == imp_info.get_dict().get('Rimp_rel')
                    ):

                        self.report('impurity_info node from input and from previous GF calculation are compatible')
                        check_consistency_imp_info = True
                    else:
                        self.report(
                            'impurity_info node from input and from previous GF calculation are NOT compatible!. '
                            'Please check your impurity_info nodes for consistency.'
                        )
                        check_consistency_imp_info = False
                except AttributeError:
                    self.report(
                        'Non default values of the impurity_info node from input and from previous '
                        "GF calculation are compatible. Default values haven't been checked"
                    )
                    check_consistency_imp_info = True
            else:
                self.report(
                    'impurity_info node from input and from previous GF calculation are NOT compatible!. '
                    'Please check your impurity_info nodes for consistency.'
                )
                check_consistency_imp_info = False
            if check_consistency_imp_info:
                imp_info = imp_info_inputnode
            else:
                raise InputValidationError('impurity_info nodes (input and GF calc) are not compatible')

        # check if host parent was KKRFLEX calculation
        hostfolder = parent_calc.outputs.retrieved
        with hostfolder.open(KkrCalculation._DEFAULT_INPUT_FILE) as fhandle:
            # use read_keywords_from_inputcard of kkrparams class
            params_host_calc = kkrparams(params_type='kkr')
            params_host_calc.read_keywords_from_inputcard(inputcard=fhandle)

        if 'RUNOPT' not in list(params_host_calc.get_dict().keys()):
            host_ok = False
        elif 'KKRFLEX' not in params_host_calc.get_dict().get('RUNOPT', []):
            host_ok = False
        else:
            host_ok = True

        if not host_ok:
            raise InputValidationError('host_Greenfunction calculation was not a KKRFLEX run')

        # extract information from Efshift host GF input node (not mandatory)
        if 'host_Greenfunction_folder_Efshift' in self.inputs:
            host_parent_Efshift = self.inputs.host_Greenfunction_folder_Efshift
            parent_calcs_Efshift = host_parent_Efshift.get_incoming(node_class=CalcJob)
            parent_calc_Efshift = parent_calcs_Efshift.first().node
            hostfolder_Efshift = parent_calc_Efshift.outputs.retrieved
        else:
            hostfolder_Efshift = None

        kkrflex_file_paths = {}
        for filename in self._ALL_KKRFLEX_FILES:
            if filename in hostfolder.list_object_names():
                kkrflex_file_paths[filename] = hostfolder
            # take tmat and green file from Fermi level overwrite directory (second GF_writeout calculation)
            if hostfolder_Efshift is not None and filename in [self._KKRFLEX_TMAT, self._KKRFLEX_GREEN]:
                if filename in hostfolder_Efshift.list_object_names():
                    kkrflex_file_paths[filename] = hostfolder_Efshift

        # extract shapes array from parameters read from inputcard
        shapes = params_host_calc.get_dict().get('<SHAPE>', None)
        if shapes is None:
            # fallback if SHAPES is not explicityl set (assume each atom has it's own shapefun
            shapes = range(1, params_host_calc.get_dict().get('NAEZ') + 1)
        elif type(shapes) == int:
            shapes = [shapes]

        # extract input structure and voro_parent to get shapefun in next step
        try:
            structure, voro_parent = VoronoiCalculation.find_parent_structure(parent_calc)
        except:
            raise InputValidationError('No structure node found from host GF parent')

        # extract shapefun path for read-in
        shapefun_path = {}
        if VoronoiCalculation._SHAPEFUN in voro_parent.outputs.retrieved.list_object_names():
            shapefun_path = voro_parent.outputs.retrieved
        else:
            shapefun_path = None

        return imp_info, kkrflex_file_paths, shapefun_path, shapes, parent_calc, params_host_calc, structure

    def _check_and_extract_input_nodes(self, tempfolder):
        """
        Extract input nodes from inputdict and check consitency of input nodes
        :param inputdict: dict of inputnodes
        :returns:
            * parameters (aiida_kkr.tools.kkr_params.kkrparams), optional: parameters of KKRimp that end up in config.cfg
            * code (KKRimpCodeNode): code of KKRimp on some machine
            * imp_info (DictNode): parameter node of the impurity information, extracted from host_parent_calc
            * kkrflex_file_paths (dict): dictionary of {filenames: absolute_path_to_file} for the kkrflex-files
            * shapfun_path (str): absolute path of the shapefunction of the host parent calculation
            * host_parent_calc (KkrCalculation): node of the parent host calculation where the kkrflex-files were created
            * impurity_potential (SinglefileData): single file data node containing the starting potential for the impurity calculation
            * parent_calc_folder (RemoteData): remote directory of a parent KKRimp calculation
        """

        # get mandatory input nodes (extract code)
        code = self.inputs.code

        # now check for optional nodes
        if 'parameters' in self.inputs:
            parameters = self.inputs.parameters
        else:
            parameters = None
        if parameters is not None:  # convert to kkrparams instance
            parameters = kkrparams(params_type='kkrimp', **parameters.get_dict())

        # get hostfiles
        imp_info, kkrflex_file_paths, shapfun_path, shapes, host_parent_calc, params_host, structure = self._get_and_verify_hostfiles(
            tempfolder
        )

        # check impurity potential or parent calculation input
        # impurity_potential
        if 'impurity_potential' in self.inputs:
            impurity_potential = self.inputs.impurity_potential
            found_imp_pot = True
        else:
            impurity_potential = None
            found_imp_pot = False
        # parent calculation folder
        if 'parent_calc_folder' in self.inputs:
            parent_calc_folder = self.inputs.parent_calc_folder
            found_parent_calc = True
        else:
            parent_calc_folder = None
            found_parent_calc = False
        # consistency checks
        if not found_parent_calc and not found_imp_pot:
            raise InputValidationError(
                'Neither impurity_potential nor parent_calc_folder specified for this calculation.\n'
                'Please provide either impurity_potential or parent_calc_folder.'
            )
        elif found_parent_calc and found_imp_pot:
            raise InputValidationError(
                'Both impurity_potential and parent_calc_folder specified for this calculation.\n'
                'Please provide one one, i.e. either impurity_potential or parent_calc_folder.'
            )

        # Done checking inputs, returning...
        return parameters, code, imp_info, kkrflex_file_paths, shapfun_path, shapes, host_parent_calc, params_host, impurity_potential, parent_calc_folder, structure

    def _extract_and_write_config(self, parent_calc_folder, params_host, parameters, tempfolder, GFhost_folder):
        """
        fill kkr params for KKRimp and write config file
        also writes kkrflex_llyfac file if Lloyd is used in the host system
        """

        runflag, params_kkrimp = self._initialize_kkrimp_params(params_host, parameters, GFhost_folder, tempfolder)

        # overwrite keys if found in parent_calc (previous KKRimp calculation)
        # here `parent_calc_folder` is the `remote` output node of the previous KKRimp calculation
        if parent_calc_folder is not None:
            parent_calc = parent_calc_folder.get_incoming().get_node_by_label('remote_folder')
            params_parent = parent_calc.get_incoming().get_node_by_label('parameters')
        else:
            params_parent = None
        if params_parent is not None:
            params_parent = kkrparams(params_type='kkrimp', **params_parent.get_dict())
            for (key, val) in params_parent.get_set_values():
                self._check_key_setting_consistency(params_kkrimp, key, val)
                params_kkrimp.set_value(key, val)

        # finally overwrite from input parameters
        if parameters is not None:
            for (key, val) in parameters.get_set_values():
                self._check_key_setting_consistency(params_kkrimp, key, val)
                params_kkrimp.set_value(key, val)

        # special run mode: calculation of Jijs
        if parameters.get_value('CALCJIJMAT') is not None and parameters.get_value('CALCJIJMAT') == 1:
            self.report('Found CALCJIJMAT=1: trigger JIJ mode which overwrites IMIX, MIXFAC, SCFSTEPS and RUNFLAGs')
            runflag = self._activate_jij_calc(runflag, params_kkrimp, GFhost_folder, tempfolder)

        # write kkrflex_angle file
        # DOES NOT WORK TOGETHER WITH JIJ mode!!!
        if 'initial_noco_angles' in self.inputs:
            self.report('Found `initial_noco_angles` input node, writing kkrflex_angle file')
            self._write_kkrflex_angle(parameters, GFhost_folder, tempfolder)

        # write config.cfg
        with tempfolder.open(self._CONFIG, u'w') as config_file:
            params_kkrimp.fill_keywords_to_inputfile(output=config_file)

    def _change_atominfo(self, imp_info, kkrflex_file_paths, tempfolder):
        """
        change kkrflex_atominfo to match impurity case
        """
        # read in atominfo file as it is written out
        with kkrflex_file_paths[KkrCalculation._KKRFLEX_ATOMINFO].open(KkrCalculation._KKRFLEX_ATOMINFO) as file:
            atominfo = file.readlines()

        #TODO implement logic to extract this info from imp_info
        replace_zatom_imp = []

        # read scoef for comparison with Rimp_rel
        scoef = []
        with tempfolder.open(KkrCalculation._SCOEF, u'r') as scoeffile:
            n_rows = len(scoeffile.readlines()) - 1
        with tempfolder.open(KkrCalculation._SCOEF, u'r') as scoeffile:
            if n_rows > 1:
                scoef = np.loadtxt(scoeffile, skiprows=1)[:, :3]
            else:
                scoef = np.loadtxt(scoeffile, skiprows=1)[:3]

        # find replaceZimp list from Zimp and Rimp_rel
        imp_info_dict = imp_info.get_dict()
        Zimp_list = imp_info_dict.get(u'Zimp')
        if type(Zimp_list) != list:
            Zimp_list = [Zimp_list]  # fast fix for cases when Zimp is not a list but a single value
        Rimp_rel_list = imp_info_dict.get(u'Rimp_rel', [[0, 0, 0]])
        self.report(f'DEBUG: Rimp_rel_list: {Rimp_rel_list}.')

        for iatom in range(len(Zimp_list)):
            rtmp = np.array(Rimp_rel_list[iatom])[:3]
            self.report(f'INFO: Rimp_rel {iatom}, {rtmp}')
            if n_rows > 1:
                diff = np.sqrt(np.sum((rtmp - scoef)**2, axis=1))
            else:
                diff = np.sqrt(np.sum((rtmp - scoef)**2, axis=0))
            Zimp = Zimp_list[iatom]
            ipos_replace = np.where(diff == diff.min())[0][0]
            replace_zatom_imp.append([ipos_replace, Zimp])

        for (iatom, zimp) in replace_zatom_imp:
            tmp = atominfo[iatom + 4].split()
            x, y, z = float(tmp[0]), float(tmp[1]), float(tmp[2])
            zatom = float(tmp[3])
            virt, remove, lmax = int(tmp[4]), int(tmp[5]), int(tmp[6])
            zatom = zimp
            tmp = u' %24.16f %24.16f %24.16f %5.2f %4i %4i %4i\n' % (x, y, z, zatom, virt, remove, lmax)
            atominfo[iatom + 4] = tmp

        # write atominfo file
        with tempfolder.open(self._KKRFLEX_ATOMINFO, u'w') as atominfofile:
            atominfofile.writelines(atominfo)

    def _get_pot_and_shape(
        self, imp_info, shapefun, shapes, impurity_potential, parent_calc_folder, tempfolder, structure
    ):
        """
        write shapefun from impurity info and host shapefun and copy imp. potential

        returns: file handle to potential file
        """

        imp_info_dict = imp_info.get_dict()

        # create scoef file
        if 'imp_cls' not in imp_info_dict:
            # this means cluster is found from parameters in imp_info

            # extract cluster settings
            Rcut = imp_info_dict.get('Rcut', None)
            hcut = imp_info_dict.get('hcut', -1.)
            cylinder_orient = imp_info_dict.get('cylinder_orient', [0., 0., 1.])
            ilayer_center = imp_info_dict.get('ilayer_center', 0)

            # now write scoef file
            self.report('Input parameters for make_scoef read in correctly!')
            with tempfolder.open(KkrCalculation._SCOEF, 'w') as scoef_file:
                make_scoef(structure, Rcut, scoef_file, hcut, cylinder_orient, ilayer_center)

        else:
            # this means the full imp cluster is given in the input

            self.report(f"Write scoef from imp_cls input! {len(imp_info_dict.get('imp_cls'))}")
            with tempfolder.open(KkrCalculation._SCOEF, 'w') as scoef_file:
                write_scoef_full_imp_cls(imp_info, scoef_file)

        # now create impurity shapefun (reads scoef file)
        with tempfolder.open(KkrCalculation._SCOEF, u'r') as scoef_file:
            with tempfolder.open(KkrimpCalculation._SHAPEFUN, u'w') as shapefun_new:
                with shapefun.open(KkrimpCalculation._SHAPEFUN, u'r') as shapefun_file:
                    shapelen = len(shapefun_file.readlines())
                    if shapelen > 1:
                        modify_potential().shapefun_from_scoef(scoef_file, shapefun_file, shapes, shapefun_new)

        # get path of tempfolder
        with tempfolder.open('.dummy', 'w') as tmpfile:
            tempfolder_path = os.path.dirname(tmpfile.name)

        # find path to input potential
        if impurity_potential is not None:
            potfile_name, potfile_folder = impurity_potential.filename, impurity_potential
        elif parent_calc_folder is not None:
            self.report(
                f'parent_calc_folder {parent_calc_folder} {parent_calc_folder.get_incoming().all_link_labels()}'
            )
            retrieved = parent_calc_folder.get_incoming(node_class=CalcJobNode
                                                        ).first().node.get_outgoing().get_node_by_label('retrieved')
            self.report(f'potfile {retrieved} {self._OUT_POTENTIAL}')

            # extract file from host's tarball (extract to tempfolder and use from there, this way the unnessesary files are deleted once submission is done)
            tar_filenames = []
            if self._FILENAME_TAR in retrieved.list_object_names():
                # get path of tarfile
                with retrieved.open(self._FILENAME_TAR) as tf:
                    tfpath = tf.name
                # extract file from tarfile of retrieved to tempfolder
                with tarfile.open(tfpath) as tf:
                    tar_filenames = [ifile.name for ifile in tf.getmembers()]
                    self.report(f'extract potfile from tarball {tar_filenames}')
                    filename = self._OUT_POTENTIAL
                    if filename in tar_filenames:
                        tf.extract(filename, tempfolder_path)  # extract to tempfolder
            else:  # otherwise copy from retrieved to tempfolder (rest of calculation needs files to be in tempfolder)
                filename = self._OUT_POTENTIAL
                if filename in retrieved.list_object_names():
                    with tempfolder.open(filename, u'w') as newfile:
                        with retrieved.open(filename, u'r') as oldfile:
                            newfile.writelines(oldfile.readlines())
                    self.report('copied potfile from retrieved to tempfolder')

            # now out_potential is in tempfolder (either copied or extracted) and can be copied from there
            potfile_name, potfile_folder = self._OUT_POTENTIAL, tempfolder
        else:
            raise InputValidationError(
                'ERROR in _get_pot_and_shape: neither impurity potential nor parent_calc_folder given!'
            )

        # copy potential to correct input name ('potential')
        with potfile_folder.open(potfile_name, u'r') as oldfile:
            with tempfolder.open(self._POTENTIAL, u'w') as newfile:
                newfile.writelines(oldfile.readlines())

        # remove old potential file if still present (saves unnecessary copying)
        if self._OUT_POTENTIAL in os.listdir(tempfolder_path):
            os.remove(os.path.join(tempfolder_path, self._OUT_POTENTIAL))
        if '.dummy' in os.listdir(tempfolder_path):
            os.remove(os.path.join(tempfolder_path, '.dummy'))

    def _get_natom(self, tempfolder):
        """Get the number of atoms in the impurity cluster from kkrflex_atominfo file"""
        with tempfolder.open(self._KKRFLEX_ATOMINFO) as file:
            atominfo = file.readlines()
        itmp = search_string('NATOM', atominfo)
        if itmp >= 0:
            natom = int(atominfo[itmp + 1].split()[0])
        else:
            raise ValueError('Could not extract NATOM value from kkrflex_atominfo')
        return natom

    def _initialize_kkrimp_params(self, params_host, parameters, GFhost_folder, tempfolder):
        """Initialize KKRimp parameters and set keys that are the same as in the host calculation"""

        # initialize kkrimp parameter set with default values
        params_kkrimp = kkrparams(
            params_type='kkrimp',
            NPAN_LOGPANELFAC=2,
            RADIUS_MIN=-1,
            NCOLL=0,
            SPINORBIT=0,
            SCFSTEPS=1,
            IMIX=0,
            MIXFAC=0.05,
            ITDBRY=20,
            BRYMIX=0.05,
            QBOUND=10**-7,
            RUNFLAG=[],
            TESTFLAG=[],
            HFIELD=[0.0, 0],
            CALCFORCE=0,
            CALCJIJMAT=0,
            CALCORBITALMOMENT=0,
            ICST=2
        )

        # keys that are being overwritten from host calculation settings
        keys_overwrite = [
            'NSPIN', 'KVREL', 'XC', 'INS', 'ICST', 'RADIUS_LOGPANELS', 'NPAN_EQ', 'NPAN_LOG', 'NCHEB', 'QBOUND'
        ]
        for key in keys_overwrite:
            if key == 'XC':
                key0 = 'KEXCOR'
            elif key == 'RADIUS_LOGPANELS':
                key0 = 'R_LOG'
            elif key == 'MIXFAC':
                key0 = 'STRMIX'
            else:
                key0 = key
            val = params_host.get_value(key0)
            if val is not None:
                params_kkrimp.set_value(key, val)

        # settings for SOC solver
        use_cheby = False
        runopts = params_host.get_value('RUNOPT')
        if runopts is not None and 'NEWSOSOL' in runopts:
            use_cheby = True
        if (
            params_host.get_value('<USE_CHEBYCHEV_SOLVER>') is not None and
            params_host.get_value('<USE_CHEBYCHEV_SOLVER>')
        ):
            use_cheby = True
        if use_cheby:
            params_kkrimp.set_multiple_values(NCOLL=1, SPINORBIT=1, CALCORBITALMOMENT=1, TESTFLAG=['tmatnew'])
        else:
            params_kkrimp.set_multiple_values(NCOLL=0, SPINORBIT=0, CALCORBITALMOMENT=0, TESTFLAG=[])

        # SOC solver but with SOC strength scaled to zero
        if use_cheby:
            # maybe deactivate SOC in KKRimp calculation
            self._set_nosoc(params_host, GFhost_folder, tempfolder)

        # extract input RUNFLAGS
        runflag = None
        if parameters is not None:
            for (key, val) in parameters.get_set_values():
                if key == 'RUNFLAG':
                    runflag = list(val)
        if runflag is None:
            runflag = []

        # special settings
        runopts = params_host.get_value('RUNOPT')
        if 'SIMULASA' in runopts or (params_kkrimp.get_value('NCOLL') > 0 and params_kkrimp.get_value('INS') == 0):
            runflag.append('SIMULASA')
        # take care of LLYsimple (i.e. Lloyd in host system)
        if 'LLOYD' in runopts:
            runflag = self._use_lloyd(runflag, GFhost_folder, tempfolder)
        # alternative to Lloyd mode: renormalization factor of energy integration weights
        if params_host.get_value('WFAC_RENORM') is not None:
            if params_host.get_value('WFAC_RENORM'):
                runflag = self._use_lloyd(runflag, GFhost_folder, tempfolder, True)

        # now set runflags
        params_kkrimp.set_value('RUNFLAG', runflag)

        return runflag, params_kkrimp

    def _set_nosoc(self, params_host, GFhost_folder, tempfolder):
        """Check if host is a noSOC calculation and then set the kkrflex_spinorbitperatom accordingly"""

        # check if noSOC mode was used in the host run
        nosoc = False
        if (params_host.get_value('<SET_CHEBY_NOSOC>') is not None and params_host.get_value('<SET_CHEBY_NOSOC>')):
            nosoc = True
        if params_host.get_value('<SOCSCL>') is not None:
            # TODO make this more flexible, now only SOC on/off is possible
            socscale = params_host.get_value('<SOCSCL>')
            try:
                for scl in socscale:
                    if scl < 1e-14:
                        nosoc = True
            except TypeError:
                if socscale < 1e-14:
                    nosoc = True

        # maybe deactivate SOC by writing the kkrflex_spinorbitperatom file
        if nosoc:
            # extract NATOM from atominfo file
            natom = self._get_natom(GFhost_folder)

            # set soc scale factor (should be integer at the moment!)
            socscale = [0 for i in range(natom)]

            # now write kkrflex_spinorbitperatom file
            with tempfolder.open(self._KKRFLEX_SOCFAC, 'w') as kkrflex_socfac:
                for socscl in socscale:
                    kkrflex_socfac.write(f' {socscl}\n')

    def _use_lloyd(self, runflag, GFhost_folder, tempfolder, wfac_renorm=False):
        """Use the LLYsimple version of KKRimp code with the average renormalization factor from the host calculation"""

        # add runflag for imp code
        runflag.append('LLYsimple')

        # find renormalization factor
        if wfac_renorm:
            # option 1: found wfac_renorm in Kkrhost parent's input parameters
            with GFhost_folder.open('inputcard') as f:
                txt = f.readlines()
            iline = search_string('WFAC_RENORM', txt)
        else:
            # option 2: host parent is a Lloyd calculation
            # extract renormalization factor and create kkrflex_llyfac file (contains one value only)
            with GFhost_folder.open('output.000.txt') as f:
                txt = f.readlines()
            iline = search_string('RENORM_LLY: Renormalization factor of total charge', txt)

        if iline >= 0:
            llyfac = txt[iline].split()[-1]
        else:
            raise ValueError('Failed to extract llyfac in KkrimpCalculation')

        # now write kkrflex_llyfac to tempfolder where later on config file is also written
        with tempfolder.open(self._KKRFLEX_LLYFAC, 'w') as f2:
            f2.writelines([llyfac])
        return runflag

    def _activate_jij_calc(self, runflag, params_kkrimp, GFhost_folder, tempfolder):
        """Adapt runoptions to use Jij calculation and set inputs for Jij run"""

        # settings in config file
        runflag.append('force_angles')
        # take care of LDA+U
        if 'settings_LDAU' in self.inputs:
            # this prevents mixing LDAU potential in between iterations
            runflag.append('freezeldau')

        # now add runflags
        params_kkrimp.set_multiple_values(IMIX=0, MIXFAC=0., SCFSTEPS=3, RUNFLAG=runflag)

        # for DOS mode add flag to writeout Jij info energy resolved (this will be retrieved if )
        testflag = params_kkrimp.get_value('TESTFLAG')
        testflag.append('Jij(E)')
        params_kkrimp.set_value('TESTFLAG', testflag)

        # extract NATOM from atominfo file
        natom = self._get_natom(GFhost_folder)

        # now write kkrflex_angle file
        with tempfolder.open(self._KKRFLEX_ANGLE, 'w') as kkrflex_angle_file:
            for istep in range(3):
                for iatom in range(natom):
                    if istep == 0:
                        kkrflex_angle_file.write(f'   0.0    0.0    1\n')
                    elif istep == 1:
                        kkrflex_angle_file.write(f'  90.0    0.0    1\n')
                    else:
                        kkrflex_angle_file.write(f'  90.0   90.0    1\n')

        return runflag

    def _write_kkrflex_angle(self, parameters, GFhost_folder, tempfolder):
        """Create the kkrflex_angle file in tempfolder"""

        # check if calculation is no Jij run
        if parameters.get_value('CALCJIJMAT') is not None and parameters.get_value('CALCJIJMAT') == 1:
            raise InputValidationError('ERROR: angles cannot be set if Jij mode is chosen!')

        # extract NATOM from atominfo file
        natom = self._get_natom(GFhost_folder)

        # extract values from input node
        thetas = self.inputs.initial_noco_angles['theta']
        if len(thetas) != natom:
            raise InputValidationError(
                'Error: `theta` list in `initial_noco_angles` input node needs to have the same length as number of atoms in the impurity cluster!'
            )
        phis = self.inputs.initial_noco_angles['phi']
        if len(phis) != natom:
            raise InputValidationError(
                'Error: `phi` list in `initial_noco_angles` input node needs to have the same length as number of atoms in the impurity cluster!'
            )
        fix_dirs = self.inputs.initial_noco_angles['fix_dir']
        if len(fix_dirs) != natom:
            raise InputValidationError(
                'Error: `fix_dir` list in `initial_noco_angles` input node needs to have the same length as number of atoms in the impurity cluster!'
            )

        # now write kkrflex_angle file
        with tempfolder.open(self._KKRFLEX_ANGLE, 'w') as kkrflex_angle_file:
            for iatom in range(natom):
                theta, phi, fix_dir = thetas[iatom], phis[iatom], fix_dirs[iatom]
                # check consistency
                if theta < 0 or theta > 180:
                    raise InputValidationError(
                        f'Error: theta value out of range (0..180): iatom={iatom}, theta={theta}'
                    )
                if phi < 0 or phi > 360:
                    raise InputValidationError(f'Error: phi value out of range (0..360): iatom={iatom}, phi={phi}')
                # convert fix_dir to integer if given as boolean
                if isinstance(fix_dir, bool):
                    fix_dir = (1 if fix_dir else 0)
                # write line
                kkrflex_angle_file.write(f'   {theta}    {phi}    {fix_dir}\n')

    def _write_kkrflex_rimpshift(self, tempfolder, parameters):
        """Create the kkrflex_rimpshift file in tempfolder for U-transformation"""

        if 'rimpshift' in self.inputs:
            # check if calculation is no Jij run
            if parameters.get_value('LATTICE_RELAX') is not None and parameters.get_value('LATTICE_RELAX') != 1:
                raise InputValidationError('ERROR: "LATTICE_RELAX" in the input parameters needs to be set to 1.')

            # extract NATOM from atominfo file
            natom = self._get_natom(tempfolder)

            # extract values from input node
            rimpshift = self.inputs.rimpshift['shifts']
            if len(rimpshift) != natom:
                raise InputValidationError(
                    'Error: `shifts` list in `rimpshift` input node needs to have the same length as number of atoms in the impurity cluster!'
                )

            # now write kkrflex_rimpshift file
            with tempfolder.open(self._KKRFLEX_RIMPSHIFT, 'w') as kkrflex_rimpshift_file:
                for iatom in range(natom):
                    shift = rimpshift[iatom]
                    # write line
                    kkrflex_rimpshift_file.write(f'   {shift[0]}    {shift[1]}    {shift[2]}\n')

    def _check_key_setting_consistency(self, params_kkrimp, key, val):
        """
        Check if key/value pair that is supposed to be set is not in conflict with previous settings of parameters in params_kkrimp
        """
        param_ok = True

        #TODO implement checks

        if not param_ok:
            raise ValueError(f'Trying to set key "{key}" with value "{val}" which is in conflict to previous settings!')

    def get_run_test_opts(self, parameters):
        """Extract run and test options from input parameters"""
        runopts = parameters.get_dict().get('RUNFLAG')
        if runopts is None:
            runopts = []
        testopts = parameters.get_dict().get('TESTFLAG')
        if testopts is None:
            testopts = []
        allopts = runopts + testopts
        return allopts

    def add_lmdos_files_to_retrieve(self, tempfolder, allopts, retrieve_list, kkrflex_file_paths):
        """Add DOS files to retrieve list"""

        if 'lmdos' in allopts or 'ldos' in allopts:
            # check if mode is Jij
            with tempfolder.open(self._CONFIG) as file:
                config = file.readlines()
            itmp = search_string('CALCJIJMAT', config)
            if itmp >= 0:
                calcjijmat = int(config[itmp].split()[-1])
            else:
                raise ValueError('Could not extract CALCJIJMAT value from config.cfg')

            # add DOS output files accordingly (names have '*' ending to catch all files for atoms
            retrieve_list.append(self._OUT_LDOS)
            retrieve_list.append(self._OUT_LDOS_INTERPOL)
            retrieve_list.append(self._OUT_LMDOS)
            retrieve_list.append(self._OUT_LMDOS_INTERPOL)
            # add Jij of E file if Jij mode
            if calcjijmat > 0:
                with kkrflex_file_paths[self._KKRFLEX_TMAT].open(self._KKRFLEX_TMAT, 'r') as f:
                    txt = []
                    for iline in range(3):
                        txt.append(f.readline())
                    nepts = int(txt[1].split()[3])
                for ie in range(1, nepts + 1):
                    retrieve_list.append(self._OUT_JIJ_OF_E)  # energy resolved values

        return retrieve_list

    def adapt_retrieve_tmatnew(self, tempfolder, allopts, retrieve_list):
        """Add out_magneticmoments and orbitalmoments files to retrieve list"""

        # extract NSPIN value
        with tempfolder.open(self._CONFIG) as file:
            config = file.readlines()
        itmp = search_string('NSPIN', config)
        if itmp >= 0:
            nspin = int(config[itmp].split()[-1])
        else:
            raise ValueError('Could not extract NSPIN value from config.cfg')

        # change retrieve list
        if 'tmatnew' in allopts and nspin > 1:
            retrieve_list.append(self._OUT_MAGNETICMOMENTS)
            with tempfolder.open(self._CONFIG) as file:
                outorb = file.readlines()
            itmp = search_string('CALCORBITALMOMENT', outorb)
            if itmp >= 0:
                calcorb = int(outorb[itmp].split()[-1])
            else:
                calcorb = 0
            if calcorb == 1:
                retrieve_list.append(self._OUT_ORBITALMOMENTS)

        return retrieve_list

    def init_ldau(self, tempfolder, retrieve_list, parent_calc_folder):
        """
        Check if settings_LDAU is in input and set up LDA+U calculation. Reuse old ldaupot of parent_folder contains a file ldaupot.
        """

        # first check if settings_LDAU is in inputs
        if 'settings_LDAU' not in self.inputs:
            # do nothing
            return retrieve_list
        else:
            # this means we need to set up LDA+U

            #TODO: check consistency of settings_LDAU

            # add ldaupot to retrieve and local copy lists
            retrieve_list.append(self._LDAUPOT)

            # add runopt LDA+U
            with tempfolder.open(self._CONFIG) as file:
                config = file.readlines()
            itmp = search_string('RUNFLAG', config)
            if itmp >= 0:
                config[itmp] = config[itmp].replace('\n', ' LDA+U \n')
                # overwrite config.cfg
                with tempfolder.open(self._CONFIG, 'w') as file:
                    file.writelines(config)
            else:
                raise ValueError('Could not find RUNFLAG in config.cfg to add LDA+U option')

            # now create ldaupot file
            self.create_or_update_ldaupot(parent_calc_folder, tempfolder)

        return retrieve_list

    def create_or_update_ldaupot(self, parent_calc_folder, tempfolder):
        """
        Writes ldaupot to tempfolder.

        If parent_calc_folder is found and it contains an onld ldaupot, we reuse the values for wldau, uldau and phi from there.
        """

        # extract Fermi energy from potential file
        with tempfolder.open(self._POTENTIAL) as potfile:
            ef_Ry = get_ef_from_potfile(potfile)

        # extract NATOM from atominfo file
        natom = self._get_natom(tempfolder)

        # get old ldaupot file
        reuse_old_ldaupot = self.get_old_ldaupot(parent_calc_folder, tempfolder)

        # settings dict (defines U, J etc.)
        ldau_settings = self.inputs.settings_LDAU.get_dict()

        if reuse_old_ldaupot:
            # reuse wldau, ildau and phi from old ldaupot file
            # Attention the first number needs to be non-zero
            txt = get_ldaupot_text(ldau_settings, ef_Ry, natom, initialize=False)
            # now we read the old file
            with tempfolder.open(self._LDAUPOT + '_old', 'r') as ldaupot_file:
                txt0 = ldaupot_file.readlines()
                # find start of wldau etc.
                ii = 0
                for line in txt0:
                    if 'wldau' in line:
                        break
                    ii += 1
                txt0 = txt0[ii:]

            #remove last line (is replace from txt0)
            txt.pop(-1)
            # put new header and old bottom together
            newtxt = txt + ['\n'] + txt0
        else:
            # initialize ldaupot file
            # Attention: here the first number needs to be 0 which triggers generating initial values in KKRimp
            newtxt = get_ldaupot_text(ldau_settings, ef_Ry, natom, initialize=True)

        # now write to file
        with tempfolder.open(self._LDAUPOT, 'w') as out_filehandle:
            out_filehandle.writelines(newtxt)

    def get_old_ldaupot(self, parent_calc_folder, tempfolder):
        """
        Copy old ldaupot from retrieved of parent or extract from tarball.
        If no parent_calc_folder is present this step is skipped.
        """

        has_ldaupot = False

        if parent_calc_folder is not None:
            retrieved = parent_calc_folder.get_incoming(node_class=CalcJobNode
                                                        ).first().node.get_outgoing().get_node_by_label('retrieved')
            # extract file from host's tarball (extract to tempfolder and use from there, this way the unnessesary files are deleted once submission is done)
            has_ldaupot = self.get_ldaupot_from_retrieved(retrieved, tempfolder)

            return has_ldaupot

    def get_remote_symlink(self, local_copy_list):
        """Check if host GF is found on remote machine and reuse from there"""
        remote_symlink_list = []

        # extract remote computer information
        code = self.inputs.code
        comp = code.computer
        # set upload dir (get the remote username and try 5 times if there was a connection error
        remote_user = get_username(comp)
        workdir = comp.get_workdir().format(username=remote_user)
        GFpath_remote = os.path.join(workdir, self._DIRNAME_GF_UPLOAD)

        self.report(f'local_copy_list: {local_copy_list}')

        # extract GF information from retrieved folder of host GF calc
        uuid_GF_calc = local_copy_list[0][0]
        GF_local_copy_info = [i for i in local_copy_list if i[1] == self._KKRFLEX_GREEN]
        TM_local_copy_info = [i for i in local_copy_list if i[1] == self._KKRFLEX_TMAT]
        if len(TM_local_copy_info) > 0:
            TM_local_copy_info = TM_local_copy_info[0]
        else:
            TM_local_copy_info = None
        if len(GF_local_copy_info) > 0:
            GF_local_copy_info = GF_local_copy_info[0]
        else:
            GF_local_copy_info = None

        # open transport to remote computer
        with comp.get_transport() as connection:
            # do this for GMAT if it was found in retreived folder, otherwise we must assume to find it on the remote
            filename = self._KKRFLEX_GREEN
            GF_remote_path = os.path.join(GFpath_remote, uuid_GF_calc, filename)
            # check if file exists on remote
            if connection.isfile(GF_remote_path):
                # remove GF from local copy list and add to remote symlink list
                if GF_local_copy_info is not None:
                    local_copy_list.remove(GF_local_copy_info)
                remote_symlink_list.append((comp.uuid, GF_remote_path, filename))

            # do the same for TMAT
            filename = self._KKRFLEX_TMAT
            TM_remote_path = os.path.join(GFpath_remote, uuid_GF_calc, filename)
            # check if file exists on remote
            if connection.isfile(TM_remote_path):
                # remove TMAT from local copy list and add to remote symlink list
                if TM_local_copy_info is not None:
                    local_copy_list.remove(TM_local_copy_info)
                remote_symlink_list.append((comp.uuid, TM_remote_path, filename))

        # print symlink and local copy list (for debugging purposes)
        self.report(f'local_copy_list: {local_copy_list}')
        self.report(f'symlink_list: {remote_symlink_list}')

        # now return updated remote_symlink and local_copy lists
        return remote_symlink_list, local_copy_list

    @classmethod
    def get_ldaupot_from_retrieved(self, retrieved, tempfolder):
        """
        Extract ldaupot from output of KKRimp retreived to tempfolder.
        The extracted file in tempfolder will be named ldaupot_old.

        returns True of ldaupot was found, otherwise returns False
        """
        tar_filenames = []
        has_ldaupot = False
        if self._FILENAME_TAR in retrieved.list_object_names():
            # get path of tempfolder
            with tempfolder.open('.dummy', 'w') as tmpfile:
                tempfolder_path = os.path.dirname(tmpfile.name)

            # get path of tarfile
            with retrieved.open(self._FILENAME_TAR) as tf:
                tfpath = tf.name

            # extract file from tarfile of retrieved to tempfolder
            with tarfile.open(tfpath) as tf:
                tar_filenames = [ifile.name for ifile in tf.getmembers()]
                if self._LDAUPOT in tar_filenames:
                    has_ldaupot = True
                    tf.extract(self._LDAUPOT, tempfolder_path)  # extract to tempfolder
            # copy to new filename
            if has_ldaupot:
                with tempfolder.open(self._LDAUPOT + '_old', u'w') as newfile:
                    with tempfolder.open(self._LDAUPOT, u'r') as oldfile:
                        newfile.writelines(oldfile.readlines())

            # remove dummy file
            if '.dummy' in os.listdir(tempfolder_path):
                os.remove(os.path.join(tempfolder_path, '.dummy'))

        else:  # otherwise copy from retrieved to tempfolder (rest of calculation needs files to be in tempfolder)
            if self._LDAUPOT in retrieved.list_object_names():
                has_ldaupot = True
                with tempfolder.open(self._LDAUPOT + '_old', u'w') as newfile:
                    with retrieved.open(self._LDAUPOT, u'r') as oldfile:
                        newfile.writelines(oldfile.readlines())

        return has_ldaupot

    def add_jij_files(self, tempfolder, retrieve_list):
        """
        check if KkrimpCalculation is in Jij mode and add OUT_JIJMAT to retrieve list if needed
        """

        # check if Jij mode is set in config file and add OUT_JIJMAT to retrieved list if needed
        with tempfolder.open(self._CONFIG) as file:
            config = file.readlines()
            itmp = search_string('CALCJIJMAT', config)
            if itmp >= 0:
                calcjijmat = int(config[itmp].split()[-1])
                if calcjijmat > 0:
                    retrieve_list.append(self._OUT_JIJMAT)
            else:
                raise ValueError('Could not extract CALCJIJMAT value from config.cfg')

        return retrieve_list



from aiida.orm import Code, CalcJobNode, Dict, Bool, RemoteData, StructureData, load_node
from aiida.engine import calcfunction, WorkChain, ToContext, if_
from aiida.plugins.factories import CalculationFactory
from aiida_kkr.calculations.kkrnano import KKRnanoCalculation
from aiida_kkr.data.strucwithpot import StrucWithPotData

__copyright__ = (u'Copyright (c), 2021, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.0.1'
__contributors__ = ('Markus Struckmann')


class KKRnanoDOSChain(WorkChain):
    """WorkChain to multiply two numbers and add a third, for testing and demonstration purposes."""
    #TODO: Add functionality for NOCO mode
    #TODO: change scfsteps to realistic numbers, e.g. 20, 100, 100
    _DEFAULT_KKRNANO_WC_PARA = {
        'coarse_mixing': {
            'scfsteps': {
                'value': 1
            },
            'imix': {
                'value': 0
            },
            'mixing': {
                'value': 0.05
            },
            'target_rms': {
                'value': 1e-02
            }
        },
        'colinear': {
            'scfsteps': {
                'value': 1
            },
            'imix': {
                'value': 6
            },
            'mixing': {
                'value': 0.01
            },
            'target_rms': {
                'value': 1e-08
            }
        },
        'SOC': {
            'scfsteps': {
                'value': 1
            },
            'imix': {
                'value': 6
            },
            'mixing': {
                'value': 0.01
            },
            'target_rms': {
                'value': 1e-05
            }
        },
        'DOS': {
            'imix': {
                'value': 6
            },
            'mixing': {
                'value': 0.01
            },
            'target_rms': {
                'value': 1e-08
            },
            'emin': {
                'value': -0.35
            },
            'emax': {
                'value': 0.35
            },
            'npnt1': {
                'value': 0
            },
            'npnt2': {
                'value': 70
            },
            'npnt3': {
                'value': 0
            },
            'tempr': {
                'value': 300
            }
        }
    }

    _OPTIONS_DEFAULT = {
        'queue_name': 'th1-2020-32',  # Queue name to submit jobs to
        'resources': {
            'num_machines': 1,
            'tot_num_mpiprocs': 15
        },  # ressources to allocate for the job
        'withmpi': True,  # execute KKR with mpi or without
        'custom_scheduler_commands': ''  # some additional scheduler commands
    }
    _OPTIONS_ADVANCED = {
        'queue_name': 'th1-2020-32',
        'max_wallclock_seconds': 72000,  # 20 h
        'resources': {
            'tot_num_mpiprocs': 15,
            'num_machines': 1
        },
        'withmpi': True
    }

    def str_mixing(self):
        builder = KKRnanoCalculation.get_builder()
        params = self.inputs.calc_parameters.get_dict()
        strparams = self.inputs.wf_parameters.get_dict()['coarse_mixing']

        #Start from parent or from strucwithpot
        if hasattr(self.inputs, 'strucwithpot') and not hasattr(self.inputs, 'parent_folder'):
            builder.strucwithpot = self.inputs.strucwithpot
        else:
            parent_folder = self.inputs.parent_folder
            builder.parent_folder = parent_folder

        for key in strparams:
            params[key] = strparams[key]


#             params['mixing']=strparams["mixing"]
#             params["imix"]=strparams["imix"]
#             params['target_rms']={"value": 1e-02}
#             params["scfsteps"]=strparams["scfsteps"]

#KKRnanoCalculation._check_input_dict(load_node('58f35be4-3ddc-4b24-947b-c3ddbb8959ac'),params)

#preparing the approptiate parameters
        params['soc'] = {'value': False}
        params['KORBIT'] = {'value': 0}

        #builder = KKRnanoCalculation.get_builder()

        builder.metadata.label = 'WC_coarse_mixing'  # pylint: disable=no-member

        builder.parameters = Dict(params)

        #builder.metadata.dry_run = True
        #machine=#
        builder.metadata.options = self.inputs.options.get_dict()  # pylint: disable=no-member

        builder.code = self.inputs.code  # orm.Code.get_from_string("KKRnanoVersuch2@iffslurm")
        str_calc = self.submit(builder)

        return ToContext(str_result=str_calc)

    def col(self):
        builder = KKRnanoCalculation.get_builder()
        params = self.inputs.calc_parameters.get_dict()
        colparams = self.inputs.wf_parameters.get_dict()['colinear']
        parent_folder = self.ctx.str_result.outputs.remote_folder

        for key in colparams:
            params[key] = colparams[key]

        #preparing the approptiate parameters
        params['soc'] = {'value': False}
        params['KORBIT'] = {'value': 0}

        #KKRnanoCalculation._check_input_dict(load_node('58f35be4-3ddc-4b24-947b-c3ddbb8959ac'),params)

        #builder = KKRnanoCalculation.get_builder()

        builder.metadata.label = 'WC_colinear'  # pylint: disable=no-member

        builder.parameters = Dict(params)

        #builder.metadata.dry_run = True
        #machine=#
        builder.metadata.options = self.inputs.options.get_dict()  # pylint: disable=no-member
        builder.parent_folder = parent_folder

        builder.code = self.inputs.code  # orm.Code.get_from_string("KKRnanoVersuch2@iffslurm")
        col_calc = self.submit(builder)

        return ToContext(col_result=col_calc)

    def SOC(self):
        builder = KKRnanoCalculation.get_builder()
        params = self.inputs.calc_parameters.get_dict()
        socparams = self.inputs.wf_parameters.get_dict()['SOC']

        if self.inputs.start_from_col.value:
            #Start from parent or from strucwithpot
            if hasattr(self.inputs, 'strucwithpot') and not hasattr(self.inputs, 'parent_folder'):
                builder.strucwithpot = self.inputs.strucwithpot
            else:
                parent_folder = self.inputs.parent_folder
                builder.parent_folder = parent_folder

        else:
            parent_folder = self.ctx.col_result.outputs.remote_folder
            builder.parent_folder = parent_folder

        for key in socparams:
            params[key] = socparams[key]

        #preparing the approptiate parameters
        params['soc'] = {'value': True}
        params['KORBIT'] = {'value': 1}

        #KKRnanoCalculation._check_input_dict(load_node('58f35be4-3ddc-4b24-947b-c3ddbb8959ac'),params)

        #builder = KKRnanoCalculation.get_builder()

        builder.metadata.label = 'WC_SOC'  # pylint: disable=no-member

        builder.parameters = Dict(params)

        #builder.metadata.dry_run = True

        builder.metadata.options = self.inputs.options.get_dict()  # pylint: disable=no-member

        builder.code = self.inputs.code  # orm.Code.get_from_string("KKRnanoVersuch2@iffslurm")
        soc_calc = self.submit(builder)

        return ToContext(soc_result=soc_calc)

    def DOS(self):

        builder = KKRnanoCalculation.get_builder()
        params = self.inputs.calc_parameters.get_dict()
        dosparams = self.inputs.wf_parameters.get_dict()['DOS']
        parent_folder = self.ctx.soc_result.outputs.remote_folder

        for key in dosparams:
            params[key] = dosparams[key]

        #preparing the approptiate parameters
        params['soc'] = {'value': True}
        params['KORBIT'] = {'value': 1}
        params['scfsteps'] = {'value': 1}
        params['npol'] = {'value': 0}

        #KKRnanoCalculation._check_input_dict(load_node('58f35be4-3ddc-4b24-947b-c3ddbb8959ac'),params)

        builder = KKRnanoCalculation.get_builder()

        builder.metadata.label = 'WC_DOS'  # pylint: disable=no-member

        builder.parameters = Dict(params)

        #builder.metadata.dry_run = True

        builder.metadata.options = self.inputs.options.get_dict()  # pylint: disable=no-member
        builder.parent_folder = parent_folder

        builder.code = self.inputs.code  # orm.Code.get_from_string("KKRnanoVersuch2@iffslurm")
        dos_calc = self.submit(builder)

        return ToContext(dos_result=dos_calc)

    #@calcfunction
    #def create_output_node(outself):

    def prepare_output(self):
        #         output_dictionary=self.ctx.soc_result.outputs.output_parameters.get_dict()
        #         if self.inputs.calculate_DOS.value:
        #             dos_dict=self.ctx.dos_result.outputs.output_parameters.get_dict()['DOS']
        #             output_dictionary['DOS']=dos_dict
        #         self.out('result', Dict(dict=output_dictionary))
        if self.inputs.calculate_DOS.value:
            output_dictionary = self.ctx.dos_result.outputs.output_parameters
        else:
            output_dictionary = self.ctx.soc_result.outputs.output_parameters
        self.out('result', output_dictionary)

    def check_input(self):
        #TODO add inputdict check like in calculation module
        inputdict = self.inputs.calc_parameters.get_dict()
        try:
            if not inputdict['NSPIND']['value'] == 2:
                return self.exit_codes.ERROR_INPUT_SPIN  # pylint: disable=no-member
        except:
            return self.exit_codes.ERROR_INPUT_SPIN  # pylint: disable=no-member

    def check_if_DOS_to_run(self):
        if self.inputs.calculate_DOS.value:
            return True
        else:
            return False

    def check_if_str_to_run(self):
        if self.inputs.start_from_col:
            return False
        else:
            return True

    def check_if_col_to_run(self):
        if self.inputs.start_from_col:
            return False
        else:
            return True

    @classmethod
    def define(cls, spec):
        """Specify inputs and outputs."""
        super().define(spec)

        # define input nodes (optional ones have required=False)
        spec.input(
            'calc_parameters',
            valid_type=Dict,
            required=True,
            help='Dict node that specifies the input calc_parameters for KKRnano (k-point density etc.)'
        )

        spec.input(
            'wf_parameters',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict=cls._DEFAULT_KKRNANO_WC_PARA),
            help='Dict node that specifies the WorkChain calc_parameters for KKRnano (thresholds, mixings, etc.)'
        )
        spec.input(
            'parent_folder',
            valid_type=RemoteData,
            required=False,
            help='Use a node that specifies a parent KKRnano or voronoi calculation'
        )
        spec.input(
            'strucwithpot',
            valid_type=StrucWithPotData,
            required=False,
            help=
            'Use a StrucWithPotData node, that contains the AiiDA structure and the shape and potential files of the individual atoms.'
        )

        spec.input('options', valid_type=Dict, required=False, default=lambda: Dict(dict=cls._OPTIONS_DEFAULT))
        spec.input(
            'calculate_DOS',
            valid_type=Bool,
            required=False,
            default=lambda: Bool(False),
            help='Trigger DOS calculation.'
        )
        spec.input(
            'start_from_col',
            valid_type=Bool,
            required=False,
            default=lambda: Bool(False),
            help='Start with SOC calculation.'
        )

        spec.input('code', valid_type=Code)

        spec.outline(
            cls.check_input,
            if_(cls.check_if_str_to_run)(cls.str_mixing),
            if_(cls.check_if_col_to_run)(cls.col), cls.SOC,
            if_(cls.check_if_DOS_to_run)(cls.DOS), cls.prepare_output
        )
        spec.output('result', valid_type=Dict)
        spec.exit_code(400, 'ERROR_INPUT_SPIN', message='Number of spin directions has to be set to 2 for SOC!')
